# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import directory

try:
    from neutron_fwaas.common import fwaas_constants
except ImportError:
    # FWaaS project no found
    from vmware_nsx.services.fwaas.common import fwaas_mocks \
        as fwaas_constants


def is_fwaas_v2_plugin_enabled():
    fwaas_plugin = directory.get_plugin(fwaas_constants.FIREWALL_V2)
    if fwaas_plugin:
        return True



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from vmware_nsx.db import db as nsx_db
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.services.fwaas.common import fwaas_callbacks_v2 as \
    com_callbacks
from vmware_nsx.services.fwaas.nsx_tv import edge_fwaas_driver_v2 as tv_driver

LOG = logging.getLogger(__name__)


class Nsxv3FwaasCallbacksV2(com_callbacks.NsxCommonv3FwaasCallbacksV2):
    """NSX-V3 RPC callbacks for Firewall As A Service - V2."""

    def __init__(self, with_rpc):
        super(Nsxv3FwaasCallbacksV2, self).__init__(with_rpc)
        # update the fwaas driver in case of TV plugin
        self.internal_driver = None
        if self.fwaas_enabled:
            if self.fwaas_driver.driver_name == tv_driver.FWAAS_DRIVER_NAME:
                self.internal_driver = self.fwaas_driver.get_T_driver()
            else:
                self.internal_driver = self.fwaas_driver

    @property
    def plugin_type(self):
        return projectpluginmap.NsxPlugins.NSX_T

    def get_port_rules(self, nsx_ls_id, fwg, plugin_rules):
        return self.internal_driver.get_port_translated_rules(
            nsx_ls_id, fwg, plugin_rules)

    def update_router_firewall(self, context, nsxlib, router_id,
                               router_interfaces, nsx_router_id, section_id,
                               from_fw=False):
        """Rewrite all the FWaaS v2 rules in the router edge firewall

        This method should be called on FWaaS updates, and on router
        interfaces changes.
        The purpose of from_fw is to differ between fw calls and other router
        calls, and if it is True - add the service router accordingly.
        """
        fw_rules = []
        with_fw = False
        # Add firewall rules per port attached to a firewall group
        for port in router_interfaces:
            nsx_ls_id, _nsx_port_id = nsx_db.get_nsx_switch_and_port_id(
                context.session, port['id'])

            # Check if this port has a firewall
            fwg = self.get_port_fwg(context, port['id'])
            if fwg:
                with_fw = True
                # Add plugin additional allow rules
                plugin_rules = self.core_plugin.get_extra_fw_rules(
                    context, router_id, port['id'])

                # add the FWaaS rules for this port
                # ingress/egress firewall rules + default ingress/egress drop
                # rule for this port
                fw_rules.extend(self.get_port_rules(nsx_ls_id, fwg,
                                                    plugin_rules))

        # add a default allow-all rule to all other traffic & ports
        fw_rules.append(self.internal_driver.get_default_backend_rule(
            section_id, allow_all=True))

        # update the backend router firewall
        exists_on_backend = self.core_plugin.verify_sr_at_backend(context,
                                                                  router_id)
        if from_fw:
            # fw action required
            if with_fw:
                # firewall exists in Neutron and not on backend - create
                if not exists_on_backend:
                    self.core_plugin.create_service_router(
                        context, router_id, update_firewall=False)
                    exists_on_backend = True
            else:
                # First, check if other services exist and use the sr
                sr_exists = self.core_plugin.service_router_has_services(
                    context, router_id)
                if not sr_exists and exists_on_backend:
                    # No other services that require service router - delete
                    self.core_plugin.delete_service_router(context, router_id)
                    exists_on_backend = False
        if exists_on_backend:
            nsxlib.firewall_section.update(section_id, rules=fw_rules)



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib.plugins import directory
from oslo_log import log as logging

from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.services.fwaas.common import v3_utils
from vmware_nsx.services.fwaas.nsx_v3 import edge_fwaas_driver_base \
    as base_driver
from vmware_nsxlib.v3 import nsx_constants as consts

LOG = logging.getLogger(__name__)
FWAAS_DRIVER_NAME = 'Fwaas V2 NSX-V3 driver'
RULE_NAME_PREFIX = 'Fwaas-'
DEFAULT_RULE_NAME = 'Default LR Layer3 Rule'


class EdgeFwaasV3DriverV2(base_driver.CommonEdgeFwaasV3Driver):
    """NSX-V3 driver for Firewall As A Service - V2."""

    def __init__(self):
        super(EdgeFwaasV3DriverV2, self).__init__(FWAAS_DRIVER_NAME)
        registry.subscribe(
            self.check_backend_version,
            resources.PROCESS, events.BEFORE_SPAWN)

    @property
    def core_plugin(self):
        """Get the NSX-V3 core plugin"""
        if not self._core_plugin:
            self._core_plugin = directory.get_plugin()
            if self._core_plugin.is_tvd_plugin():
                self._core_plugin = self._core_plugin.get_plugin_by_type(
                    projectpluginmap.NsxPlugins.NSX_T)
                if not self._core_plugin:
                    # The nsx-t plugin was not initialized
                    return
            # make sure plugin init was completed
            if not self._core_plugin.init_is_complete:
                self._core_plugin.init_complete(None, None, {})
        return self._core_plugin

    @property
    def nsxlib(self):
        return self.core_plugin.nsxlib

    @property
    def nsx_firewall(self):
        return self.nsxlib.firewall_section

    @property
    def nsx_router(self):
        return self.nsxlib.logical_router

    def check_backend_version(self, resource, event, trigger, payload=None):
        if (self.core_plugin and
            not self.nsxlib.feature_supported(consts.FEATURE_ROUTER_FIREWALL)):
            # router firewall is not supported
            LOG.warning("FWaaS is not supported by the NSX backend (version "
                        "%s): Router firewall is not supported",
                        self.nsxlib.get_version())
            self.backend_support = False

    def _translate_cidr(self, cidr, fwaas_rule_id):
        # Validate that this is a legal & supported ipv4 / ipv6 cidr
        error_msg = (_("Unsupported FWAAS cidr %(cidr)s for rule %(id)s") % {
                     'cidr': cidr, 'id': fwaas_rule_id})
        net = netaddr.IPNetwork(cidr)
        if net.version == 4:
            if cidr.startswith('0.0.0.0'):
                # Treat as ANY and just log warning
                LOG.warning(error_msg)
                return
            if net.prefixlen == 0:
                LOG.error(error_msg)
                raise self.driver_exception(driver=self.driver_name)
        elif net.version == 6:
            if str(net.ip) == "::" or net.prefixlen == 0:
                LOG.error(error_msg)
                raise self.driver_exception(driver=self.driver_name)
        else:
            LOG.error(error_msg)
            raise self.driver_exception(driver=self.driver_name)

        return self.nsx_firewall.get_ip_cidr_reference(
            cidr,
            consts.IPV6 if net.version == 6 else consts.IPV4)

    def translate_addresses_to_target(self, cidrs, plugin_type,
                                      fwaas_rule_id=None):
        translated_cidrs = []
        for ip in cidrs:
            res = self._translate_cidr(ip, fwaas_rule_id)
            if res:
                translated_cidrs.append(res)
        return translated_cidrs

    def _translate_services(self, fwaas_rule):
        l4_protocol = v3_utils.translate_fw_rule_protocol(
            fwaas_rule['protocol'])
        if l4_protocol in [consts.TCP, consts.UDP]:
            source_ports = []
            destination_ports = []
            if fwaas_rule.get('source_port'):
                source_ports = v3_utils.translate_fw_rule_ports(
                    fwaas_rule['source_port'])
            if fwaas_rule.get('destination_port'):
                destination_ports = v3_utils.translate_fw_rule_ports(
                    fwaas_rule['destination_port'])

            return [self.nsx_firewall.get_nsservice(
                consts.L4_PORT_SET_NSSERVICE,
                l4_protocol=l4_protocol,
                source_ports=source_ports,
                destination_ports=destination_ports)]
        elif l4_protocol == consts.ICMPV4:
            # Add both icmp v4 & v6 services
            return [
                self.nsx_firewall.get_nsservice(
                    consts.ICMP_TYPE_NSSERVICE,
                    protocol=consts.ICMPV4),
                self.nsx_firewall.get_nsservice(
                    consts.ICMP_TYPE_NSSERVICE,
                    protocol=consts.ICMPV6),
            ]

    def _translate_rules(self, fwaas_rules, replace_src=None,
                         replace_dest=None, logged=False):
        translated_rules = []
        for rule in fwaas_rules:
            nsx_rule = {}
            if not rule['enabled']:
                # skip disabled rules
                continue
            # Make sure the rule has a name, and it starts with the prefix
            # (backend max name length is 255)
            if rule.get('name'):
                name = RULE_NAME_PREFIX + rule['name']
            else:
                name = RULE_NAME_PREFIX + rule['id']
            nsx_rule['display_name'] = name[:255]
            if rule.get('description'):
                nsx_rule['notes'] = rule['description']
            nsx_rule['action'] = v3_utils.translate_fw_rule_action(
                rule['action'], rule['id'])
            if not nsx_rule['action']:
                raise self.driver_exception(driver=self.driver_name)

            if (rule.get('destination_ip_address') and
                not rule['destination_ip_address'].startswith('0.0.0.0')):
                nsx_rule['destinations'] = self.translate_addresses_to_target(
                    [rule['destination_ip_address']], rule['id'])
            elif replace_dest:
                # set this value as the destination logical switch
                # (only if no dest IP)
                nsx_rule['destinations'] = [{'target_type': 'LogicalSwitch',
                                             'target_id': replace_dest}]
            if (rule.get('source_ip_address') and
                not rule['source_ip_address'].startswith('0.0.0.0')):
                nsx_rule['sources'] = self.translate_addresses_to_target(
                    [rule['source_ip_address']], rule['id'])
            elif replace_src:
                # set this value as the source logical switch,
                # (only if no source IP)
                nsx_rule['sources'] = [{'target_type': 'LogicalSwitch',
                                        'target_id': replace_src}]
            if rule.get('protocol'):
                nsx_rule['services'] = self._translate_services(rule)
            if logged:
                nsx_rule['logged'] = logged
            # Set rule direction
            if replace_src:
                nsx_rule['direction'] = 'OUT'
            elif replace_dest:
                nsx_rule['direction'] = 'IN'
            translated_rules.append(nsx_rule)

        return translated_rules

    def validate_backend_version(self):
        # prevent firewall actions if the backend does not support it
        if not self.backend_support:
            LOG.error("The NSX backend does not support router firewall")
            raise self.driver_exception(driver=self.driver_name)

    def get_default_backend_rule(self, section_id, allow_all=True):
        # Add default allow all rule
        old_default_rule = self.nsx_firewall.get_default_rule(
            section_id)
        return {
            'display_name': DEFAULT_RULE_NAME,
            'action': (consts.FW_ACTION_ALLOW if allow_all
                       else consts.FW_ACTION_DROP),
            'is_default': True,
            'id': old_default_rule['id'] if old_default_rule else 0}

    def get_port_translated_rules(self, nsx_ls_id, firewall_group,
                                  plugin_rules):
        """Return the list of translated rules per port"""
        port_rules = []
        # TODO(asarfaty): get this value from the firewall group extensions
        logged = False
        # Add the firewall group ingress/egress rules only if the fw is up
        if firewall_group['admin_state_up']:
            port_rules.extend(self._translate_rules(
                firewall_group['ingress_rule_list'],
                replace_dest=nsx_ls_id,
                logged=logged))
            port_rules.extend(self._translate_rules(
                firewall_group['egress_rule_list'],
                replace_src=nsx_ls_id,
                logged=logged))

        # Add the per-port plugin rules
        if plugin_rules and isinstance(plugin_rules, list):
            port_rules.extend(plugin_rules)

        # Add ingress/egress block rules for this port
        port_rules.extend([
            {'display_name': "Block port ingress",
             'action': consts.FW_ACTION_DROP,
             'destinations': [{'target_type': 'LogicalSwitch',
                               'target_id': nsx_ls_id}],
             'direction': 'IN'},
            {'display_name': "Block port egress",
             'action': consts.FW_ACTION_DROP,
             'sources': [{'target_type': 'LogicalSwitch',
                          'target_id': nsx_ls_id}],
             'direction': 'OUT'}])

        return port_rules



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas.nsx_p.implementation import lb_utils
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)


class EdgeL7PolicyManagerFromDict(base_mgr.NsxpLoadbalancerBaseManager):
    def create(self, context, policy, completor):
        vs_client = self.core_plugin.nsxpolicy.load_balancer.virtual_server
        policy_name = utils.get_name_and_uuid(policy['name'] or 'policy',
                                              policy['id'])
        rule_body = lb_utils.convert_l7policy_to_lb_rule(
            self.core_plugin.nsxpolicy, policy)
        try:
            position = policy.get('position', 0) - 1
            vs_client.add_lb_rule(policy['listener_id'],
                                  name=policy_name,
                                  position=position,
                                  **rule_body)
        except nsxlib_exc.ManagerError:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error('Failed to add rule %(rule)% to virtual server '
                          '%(vs)s at NSX backend',
                          {'rule': policy['id'], 'vs': policy['listener_id']})

        completor(success=True)

    def update(self, context, old_policy, new_policy, completor):
        vs_client = self.core_plugin.nsxpolicy.load_balancer.virtual_server
        policy_name = utils.get_name_and_uuid(old_policy['name'] or 'policy',
                                              old_policy['id'])
        rule_body = lb_utils.convert_l7policy_to_lb_rule(
            self.core_plugin.nsxpolicy, new_policy)
        try:
            vs_client.update_lb_rule(
                new_policy['listener_id'],
                policy_name,
                position=new_policy.get('position', 0) - 1,
                **rule_body)

        except Exception as e:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error('Failed to update L7policy %(policy)s: '
                          '%(err)s', {'policy': old_policy['id'], 'err': e})
        completor(success=True)

    def delete(self, context, policy, completor):
        vs_client = self.core_plugin.nsxpolicy.load_balancer.virtual_server
        policy_name = utils.get_name_and_uuid(policy['name'] or 'policy',
                                              policy['id'])
        try:
            vs_client.remove_lb_rule(policy['listener_id'],
                                     policy_name)
        except nsxlib_exc.ResourceNotFound:
            pass
        except nsxlib_exc.ManagerError:
            completor(success=False)
            msg = (_('Failed to delete L7 policy: %(policy)s') %
                   {'policy': policy['id']})
            raise n_exc.BadRequest(resource='lbaas-l7policy', msg=msg)
        completor(success=True)

    def delete_cascade(self, context, policy, completor):
        self.delete(context, policy, completor)



# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def lb_hm_obj_to_dict(hm):
    # Translate the LBaaS HM to a dictionary skipping the pool object to avoid
    # recursions
    hm_dict = hm.to_dict(pool=False)
    # Translate the pool separately without it's internal objects
    if hm.pool:
        hm_dict['pool'] = lb_pool_obj_to_dict(hm.pool, with_listeners=False)
    return hm_dict


def lb_listener_obj_to_dict(listener):
    # Translate the LBaaS listener to a dictionary skipping the some objects
    # to avoid recursions
    listener_dict = listener.to_dict(loadbalancer=False, default_pool=False)

    # Translate the default pool separately without it's internal objects
    if listener.default_pool:
        listener_dict['default_pool'] = lb_pool_obj_to_dict(
            listener.default_pool, with_listeners=False)
    else:
        listener_dict['default_pool'] = None

    if listener.loadbalancer:
        listener_dict['loadbalancer'] = lb_loadbalancer_obj_to_dict(
            listener.loadbalancer)
    else:
        listener_dict['loadbalancer'] = None
    return listener_dict


def lb_pool_obj_to_dict(pool, with_listeners=True):
    # Translate the LBaaS pool to a dictionary skipping the some objects
    # to avoid recursions
    pool_dict = pool.to_dict(listeners=False, listener=False)
    if with_listeners:
        # Translate the listener/s separately without it's internal objects
        if pool.listener:
            pool_dict['listener'] = lb_listener_obj_to_dict(pool.listener)
        else:
            pool_dict['listener'] = None
        pool_dict['listeners'] = []
        if pool.listeners:
            for listener in pool.listeners:
                pool_dict['listeners'].append(
                    lb_listener_obj_to_dict(listener))
    return pool_dict


def lb_loadbalancer_obj_to_dict(loadbalancer):
    return loadbalancer.to_dict()


def lb_member_obj_to_dict(member):
    # Translate the LBaaS member to a dictionary skipping the some objects
    # to avoid recursions
    member_dict = member.to_dict(pool=False)
    # Add the pool dictionary (with its listeners and loadbalancer)
    if member.pool:
        member_dict['pool'] = lb_pool_obj_to_dict(member.pool)
    else:
        member_dict['pool'] = None
    return member_dict


def lb_l7policy_obj_to_dict(l7policy):
    # Translate the LBaaS L7 policy to a dictionary skipping the some objects
    # to avoid recursions
    l7policy_dict = l7policy.to_dict(listener=False, rules=False)
    # Add the listener dictionary
    if l7policy.listener:
        l7policy_dict['listener'] = lb_listener_obj_to_dict(l7policy.listener)
    else:
        l7policy_dict['listener'] = None
    # Add the rules
    l7policy_dict['rules'] = []
    if l7policy.rules:
        for rule in l7policy.rules:
            l7policy_dict['rules'].append(
                lb_l7rule_obj_to_dict(rule, with_policy=False))

    return l7policy_dict


def lb_l7rule_obj_to_dict(l7rule, with_policy=True):
    # Translate the LBaaS L7 rule to a dictionary skipping the some objects
    # to avoid recursions
    l7rule_dict = l7rule.to_dict(policy=False)
    # Add the policy dictionary
    if with_policy:
        l7rule_dict['policy'] = lb_l7policy_obj_to_dict(l7rule.policy)
    else:
        l7rule_dict['policy'] = None
    return l7rule_dict



# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple
import contextlib

import mock
from oslo_utils import uuidutils

from neutron.db import l3_db
from neutron.db.models import l3 as l3_models
from neutron_lib.api.definitions import external_net as extnet_apidef
from neutron_lib import context as n_ctx
from neutron_lib.plugins import directory
from neutron_vpnaas.db.vpn import vpn_models  # noqa
from neutron_vpnaas.tests import base

from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.services.vpnaas.nsxp import ipsec_driver
from vmware_nsx.services.vpnaas.nsxp import ipsec_validator
from vmware_nsx.tests.unit.nsx_p import test_plugin

_uuid = uuidutils.generate_uuid

FAKE_TENANT = _uuid()
FAKE_ROUTER_ID = "aaaaaa-bbbbb-ccc"
FAKE_ROUTER = {'id': FAKE_ROUTER_ID,
               'name': 'fake router',
               'project_id': FAKE_TENANT,
               'admin_state_up': True,
               'status': 'ACTIVE',
               'gw_port_id': _uuid(),
               'enable_snat': False,
               l3_db.EXTERNAL_GW_INFO: {'network_id': _uuid()}}
FAKE_SUBNET_ID = _uuid()
FAKE_SUBNET = {'cidr': '1.1.1.0/24', 'id': FAKE_SUBNET_ID}
FAKE_VPNSERVICE_ID = _uuid()
FAKE_VPNSERVICE = {'id': FAKE_VPNSERVICE_ID,
                   'name': 'vpn_service',
                   'description': 'dummy',
                   'router': FAKE_ROUTER,
                   'router_id': FAKE_ROUTER_ID,
                   'subnet': FAKE_SUBNET,
                   'subnet_id': FAKE_SUBNET_ID,
                   'project_id': FAKE_TENANT,
                   'external_v4_ip': '1.1.1.1',
                   'admin_state_up': True}
FAKE_IKE_POLICY_ID = _uuid()
FAKE_IKE_POLICY = {'id': FAKE_IKE_POLICY_ID,
                   'name': 'ike_dummy',
                   'description': 'ike_dummy',
                   'auth_algorithm': 'sha1',
                   'encryption_algorithm': 'aes-128',
                   'phase1_negotiation_mode': 'main',
                   'lifetime': {
                       'units': 'seconds',
                       'value': 3600},
                   'ike_version': 'v1',
                   'pfs': 'group14',
                   'project_id': FAKE_TENANT}
FAKE_IPSEC_POLICY_ID = _uuid()
FAKE_IPSEC_POLICY = {'id': FAKE_IPSEC_POLICY_ID,
                     'name': 'ipsec_dummy',
                     'description': 'myipsecpolicy1',
                     'auth_algorithm': 'sha1',
                     'encryption_algorithm': 'aes-128',
                     'encapsulation_mode': 'tunnel',
                     'lifetime': {
                         'units': 'seconds',
                         'value': 3600},
                     'transform_protocol': 'esp',
                     'pfs': 'group14',
                     'project_id': FAKE_TENANT}
FAKE_IPSEC_CONNECTION_ID = _uuid()
FAKE_IPSEC_CONNECTION = {'vpnservice_id': FAKE_VPNSERVICE_ID,
                         'ikepolicy_id': FAKE_IKE_POLICY_ID,
                         'ipsecpolicy_id': FAKE_IPSEC_POLICY_ID,
                         'name': 'VPN connection',
                         'description': 'VPN connection',
                         'id': FAKE_IPSEC_CONNECTION_ID,
                         'peer_address': '192.168.1.10',
                         'peer_id': '192.168.1.10',
                         'peer_cidrs': '192.168.1.0/24',
                         'mtu': 1500,
                         'psk': 'abcd',
                         'initiator': 'bi-directional',
                         'dpd': {
                             'action': 'hold',
                             'interval': 30,
                             'timeout': 120},
                         'admin_state_up': True,
                         'project_id': FAKE_TENANT}
FAKE_NEW_CONNECTION = {'vpnservice_id': FAKE_VPNSERVICE_ID,
                       'ikepolicy_id': FAKE_IKE_POLICY_ID,
                       'ipsecpolicy_id': FAKE_IPSEC_POLICY_ID,
                       'name': 'VPN connection',
                       'description': 'VPN connection',
                       'id': FAKE_IPSEC_CONNECTION_ID,
                       'peer_address': '192.168.1.10',
                       'peer_id': '192.168.1.10',
                       'peer_cidrs': '192.168.2.0/24',
                       'mtu': 1500,
                       'psk': 'abcd',
                       'initiator': 'bi-directional',
                       'dpd': {
                           'action': 'hold',
                           'interval': 30,
                           'timeout': 120},
                       'admin_state_up': True,
                       'project_id': FAKE_TENANT}
FAKE_VPNSERVICE_NO_SUB = {'id': FAKE_VPNSERVICE_ID,
                          'name': 'vpn_service',
                          'description': 'dummy',
                          'router': FAKE_ROUTER,
                          'router_id': FAKE_ROUTER_ID,
                          'project_id': FAKE_TENANT,
                          'external_v4_ip': '1.1.1.1',
                          'admin_state_up': True}
FAKE_ENDPOINTS_CONNECTION = {'vpnservice_id': FAKE_VPNSERVICE_ID,
                             'ikepolicy_id': FAKE_IKE_POLICY_ID,
                             'ipsecpolicy_id': FAKE_IPSEC_POLICY_ID,
                             'name': 'VPN connection',
                             'description': 'VPN connection',
                             'id': FAKE_IPSEC_CONNECTION_ID,
                             'peer_address': '192.168.1.10',
                             'peer_id': '192.168.1.10',
                             'peer_ep_group_id': 'cidr_ep',
                             'local_ep_group_id': 'subnet_ep',
                             'mtu': 1500,
                             'psk': 'abcd',
                             'initiator': 'bi-directional',
                             'dpd': {
                                 'action': 'hold',
                                 'interval': 30,
                                 'timeout': 120},
                             'admin_state_up': True,
                             'project_id': FAKE_TENANT}


class TestDriverValidation(base.BaseTestCase):

    def setUp(self):
        super(TestDriverValidation, self).setUp()
        self.context = n_ctx.Context('some_user', 'some_tenant')
        self.service_plugin = mock.Mock()
        driver = mock.Mock()
        driver.service_plugin = self.service_plugin
        with mock.patch("neutron_lib.plugins.directory.get_plugin"):
            self.validator = ipsec_validator.IPsecNsxPValidator(driver)
            self.validator._l3_plugin = mock.Mock()
            self.validator._core_plugin = mock.Mock()

        self.vpn_service = {'router_id': 'dummy_router',
                            'subnet_id': 'dummy_subnet'}
        self.peer_address = '10.10.10.10'
        self.peer_cidr = '10.10.11.0/20'

    def _test_lifetime_not_in_seconds(self, validation_func):
        policy_info = {'lifetime': {'units': 'kilobytes', 'value': 1000}}
        self.assertRaises(nsx_exc.NsxVpnValidationError,
                          validation_func,
                          self.context, policy_info)

    def test_ike_lifetime_not_in_seconds(self):
        self._test_lifetime_not_in_seconds(
            self.validator.validate_ike_policy)

    def test_ipsec_lifetime_not_in_seconds(self):
        self._test_lifetime_not_in_seconds(
            self.validator.validate_ipsec_policy)

    def _test_lifetime_seconds_values_at_limits(self, validation_func):
        policy_info = {'lifetime': {'units': 'seconds', 'value': 21600}}
        validation_func(self.context, policy_info)
        policy_info = {'lifetime': {'units': 'seconds', 'value': 86400}}
        validation_func(self.context, policy_info)

        policy_info = {'lifetime': {'units': 'seconds', 'value': 10}}
        self.assertRaises(nsx_exc.NsxVpnValidationError,
                          validation_func,
                          self.context, policy_info)

    def test_ike_lifetime_seconds_values_at_limits(self):
        self._test_lifetime_seconds_values_at_limits(
            self.validator.validate_ike_policy)

    def test_ipsec_lifetime_seconds_values_at_limits(self):
        self._test_lifetime_seconds_values_at_limits(
            self.validator.validate_ipsec_policy)

    def _test_auth_algorithm(self, validation_func):
        auth_algorithm = {'auth_algorithm': 'sha384'}
        validation_func(self.context, auth_algorithm)

        auth_algorithm = {'auth_algorithm': 'sha512'}
        validation_func(self.context, auth_algorithm)

        auth_algorithm = {'auth_algorithm': 'sha1'}
        validation_func(self.context, auth_algorithm)

        auth_algorithm = {'auth_algorithm': 'sha256'}
        validation_func(self.context, auth_algorithm)

    def test_ipsec_auth_algorithm(self):
        self._test_auth_algorithm(self.validator.validate_ipsec_policy)

    def test_ike_auth_algorithm(self):
        self._test_auth_algorithm(self.validator.validate_ike_policy)

    def _test_encryption_algorithm(self, validation_func):
        auth_algorithm = {'encryption_algorithm': 'aes-192'}
        self.assertRaises(nsx_exc.NsxVpnValidationError,
                          validation_func,
                          self.context, auth_algorithm)

        auth_algorithm = {'encryption_algorithm': 'aes-128'}
        validation_func(self.context, auth_algorithm)

        auth_algorithm = {'encryption_algorithm': 'aes-256'}
        validation_func(self.context, auth_algorithm)

    def test_ipsec_encryption_algorithm(self):
        self._test_encryption_algorithm(self.validator.validate_ipsec_policy)

    def test_ike_encryption_algorithm(self):
        self._test_encryption_algorithm(self.validator.validate_ike_policy)

    def test_ike_negotiation_mode(self):
        policy_info = {'phase1-negotiation-mode': 'aggressive'}
        self.assertRaises(nsx_exc.NsxVpnValidationError,
                          self.validator.validate_ike_policy,
                          self.context, policy_info)

        policy_info = {'phase1-negotiation-mode': 'main'}
        self.validator.validate_ike_policy(self.context, policy_info)

    def _test_pfs(self, validation_func):
        policy_info = {'pfs': 'group15'}
        self.assertRaises(nsx_exc.NsxVpnValidationError,
                          validation_func,
                          self.context, policy_info)

        policy_info = {'pfs': 'group14'}
        validation_func(self.context, policy_info)

    def test_ipsec_pfs(self):
        self._test_pfs(self.validator.validate_ipsec_policy)

    def test_ike_pfs(self):
        self._test_pfs(self.validator.validate_ike_policy)

    def test_ipsec_encap_mode(self):
        policy_info = {'encapsulation_mode': 'transport'}
        self.assertRaises(nsx_exc.NsxVpnValidationError,
                          self.validator.validate_ipsec_policy,
                          self.context, policy_info)

        policy_info = {'encapsulation_mode': 'tunnel'}
        self.validator.validate_ipsec_policy(self.context, policy_info)

    def test_ipsec_transform_protocol(self):
        policy_info = {'transform_protocol': 'ah'}
        self.assertRaises(nsx_exc.NsxVpnValidationError,
                          self.validator.validate_ipsec_policy,
                          self.context, policy_info)

        policy_info = {'transform_protocol': 'esp'}
        self.validator.validate_ipsec_policy(self.context, policy_info)

    def test_vpn_service_validation(self):
        db_router = l3_models.Router()
        nsx_router = {'ha_mode': 'ACITVE_ACTIVE'}
        db_router.enable_snat = False
        with mock.patch.object(self.validator.nsxpolicy.tier0, 'get',
                               return_value=nsx_router):
            self.assertRaises(nsx_exc.NsxVpnValidationError,
                              self.validator.validate_vpnservice,
                              self.context, self.vpn_service)

        nsx_router = {'ha_mode': 'ACTIVE_STANDBY'}
        db_router.enable_snat = True
        with mock.patch.object(self.validator.nsxpolicy.tier0, 'get',
                               return_value=nsx_router),\
            mock.patch.object(self.validator._core_plugin, '_get_router',
                              return_value=db_router):
            self.validator.validate_vpnservice(self.context, self.vpn_service)

        nsx_router = {'ha_mode': 'ACTIVE_STANDBY'}
        db_router.enable_snat = False
        with mock.patch.object(self.validator.nsxpolicy.tier0, 'get',
                               return_value=nsx_router),\
            mock.patch.object(self.validator._core_plugin, '_get_router',
                              return_value=db_router):
            self.validator.validate_vpnservice(self.context, self.vpn_service)

        nsx_router = {'ha_mode': 'ACTIVE_STANDBY'}
        db_router.enable_snat = False
        vpn_service_no_subnet = {'router_id': 'dummy_router',
                                 'subnet_id': None}
        with mock.patch.object(self.validator.nsxpolicy.tier0, 'get',
                               return_value=nsx_router),\
            mock.patch.object(self.validator._core_plugin, '_get_router',
                              return_value=db_router):
            self.validator.validate_vpnservice(
                self.context, vpn_service_no_subnet)

    def _test_conn_validation(self, conn_params=None, success=True,
                              connections=None, service_subnets=None,
                              router_subnets=None):
        if connections is None:
            connections = []
        if router_subnets is None:
            router_subnets = []

        def mock_get_routers(context, filters=None, fields=None):
            return [{'id': 'no-snat',
                     'external_gateway_info': {'enable_snat': False}}]

        def mock_get_service(context, service_id):
            if service_subnets:
                # option to give the test a different subnet per service
                subnet_cidr = service_subnets[int(service_id) - 1]
            else:
                subnet_cidr = '5.5.5.0/2%s' % service_id
            return {'id': service_id,
                    'router_id': service_id,
                    'subnet_id': 'dummy_subnet',
                    'external_v4_ip': '1.1.1.%s' % service_id,
                    'subnet': {'id': 'dummy_subnet',
                               'cidr': subnet_cidr}}

        def mock_get_connections(context, filters=None, fields=None):
            if filters and 'peer_address' in filters:
                return [conn for conn in connections
                        if conn['peer_address'] == filters['peer_address'][0]]
            else:
                return connections

        with mock.patch.object(self.validator.vpn_plugin, '_get_vpnservice',
                               side_effect=mock_get_service),\
            mock.patch.object(self.validator._core_plugin, 'get_routers',
                              side_effect=mock_get_routers),\
            mock.patch.object(self.validator._core_plugin,
                              '_find_router_subnets_cidrs',
                              return_value=router_subnets),\
            mock.patch.object(self.validator.vpn_plugin,
                              'get_ipsec_site_connections',
                              side_effect=mock_get_connections):
            ipsec_sitecon = {'id': '1',
                             'vpnservice_id': '1',
                             'mtu': 1500,
                             'peer_address': self.peer_address,
                             'peer_cidrs': [self.peer_cidr]}
            if conn_params:
                ipsec_sitecon.update(conn_params)
            if success:
                self.validator.validate_ipsec_site_connection(
                    self.context, ipsec_sitecon)
            else:
                self.assertRaises(
                    nsx_exc.NsxVpnValidationError,
                    self.validator.validate_ipsec_site_connection,
                    self.context, ipsec_sitecon)

    def test_dpd_validation(self):
        params = {'dpd': {'action': 'hold',
                          'timeout': 120}}
        self._test_conn_validation(conn_params=params, success=True)

        params = {'dpd': {'action': 'clear',
                          'timeout': 120}}
        self._test_conn_validation(conn_params=params, success=False)

        params = {'dpd': {'action': 'hold',
                          'timeout': 2}}
        self._test_conn_validation(conn_params=params, success=False)

    def test_check_unique_addresses(self):
        # this test runs with non-overlapping local subnets on
        # different routers
        subnets = ['5.5.5.0/20', '6.6.6.0/20']

        # same service/router gw & peer address - should fail
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '1',
                        'peer_address': self.peer_address,
                        'peer_cidrs': [self.peer_cidr]}]
        self._test_conn_validation(success=False,
                                   connections=connections,
                                   service_subnets=subnets)

        # different service/router gw - ok
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '2',
                        'peer_address': self.peer_address,
                        'peer_cidrs': ['6.6.6.6']}]
        self._test_conn_validation(success=True,
                                   connections=connections,
                                   service_subnets=subnets)

        # different peer address - ok
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '1',
                        'peer_address': '7.7.7.1',
                        'peer_cidrs': ['7.7.7.7']}]
        self._test_conn_validation(success=True,
                                   connections=connections,
                                   service_subnets=subnets)

        # ignoring non-active connections
        connections = [{'id': '2',
                        'status': 'ERROR',
                        'vpnservice_id': '1',
                        'peer_address': self.peer_address,
                        'peer_cidrs': [self.peer_cidr]}]
        self._test_conn_validation(success=True,
                                   connections=connections,
                                   service_subnets=subnets)

    def test_overlapping_rules(self):
        # peer-cidr overlapping with new one, same subnet - should fail
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '1',
                        'peer_address': '9.9.9.9',
                        'peer_cidrs': ['10.10.11.1/19']}]
        self._test_conn_validation(success=False,
                                   connections=connections)

        # same peer-cidr, overlapping subnets - should fail
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '2',
                        'peer_address': '9.9.9.9',
                        'peer_cidrs': [self.peer_cidr]}]
        self._test_conn_validation(success=False,
                                   connections=connections)

        # non overlapping peer-cidr, same subnet - ok
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '1',
                        'peer_address': '7.7.7.1',
                        'peer_cidrs': ['7.7.7.7']}]
        self._test_conn_validation(success=True,
                                   connections=connections)

        # ignoring non-active connections
        connections = [{'id': '2',
                        'status': 'ERROR',
                        'vpnservice_id': '1',
                        'peer_address': '9.9.9.9',
                        'peer_cidrs': ['10.10.11.1/19']}]
        self._test_conn_validation(success=True,
                                   connections=connections)

    def test_advertisment(self):
        # different routers, same subnet - should fail
        subnets = ['5.5.5.0/20', '5.5.5.0/20']
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '2',
                        'peer_address': self.peer_address,
                        'peer_cidrs': ['6.6.6.6']}]
        self._test_conn_validation(success=False,
                                   connections=connections,
                                   service_subnets=subnets)

        # different routers, overlapping subnet - should fail
        subnets = ['5.5.5.0/20', '5.5.5.0/21']
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '2',
                        'peer_address': self.peer_address,
                        'peer_cidrs': ['6.6.6.6']}]
        self._test_conn_validation(success=False,
                                   connections=connections,
                                   service_subnets=subnets)

        # different routers, non overlapping subnet - ok
        subnets = ['5.5.5.0/20', '50.5.5.0/21']
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '2',
                        'peer_address': self.peer_address,
                        'peer_cidrs': ['6.6.6.6']}]
        self._test_conn_validation(success=True,
                                   connections=connections,
                                   service_subnets=subnets)

        # no-snat router with overlapping subnet to the service subnet - fail
        subnets = ['5.5.5.0/21', '1.1.1.0/20']
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '2',
                        'peer_address': self.peer_address,
                        'peer_cidrs': ['6.6.6.6']}]
        self._test_conn_validation(success=False,
                                   connections=connections,
                                   router_subnets=subnets)

        # no-snat router with non overlapping subnet to the service subnet - ok
        service_subnets = ['5.5.5.0/20', '6.6.6.0/20']
        router_subnets = ['50.5.5.0/21', '1.1.1.0/20']
        connections = [{'id': '2',
                        'status': 'ACTIVE',
                        'vpnservice_id': '2',
                        'peer_address': self.peer_address,
                        'peer_cidrs': ['6.6.6.6']}]
        self._test_conn_validation(success=True,
                                   connections=connections,
                                   service_subnets=service_subnets,
                                   router_subnets=router_subnets)


class TestVpnaasDriver(test_plugin.NsxPPluginTestCaseMixin):

    def setUp(self):
        super(TestVpnaasDriver, self).setUp()
        self.context = n_ctx.get_admin_context()
        self.service_plugin = mock.Mock()
        self.validator = mock.Mock()
        self.driver = ipsec_driver.NSXpIPsecVpnDriver(self.service_plugin)
        self.plugin = directory.get_plugin()
        self.policy_vpn = self.plugin.nsxpolicy.ipsec_vpn
        self.l3plugin = self.plugin

        mock.patch.object(self.plugin,
                          'service_router_has_loadbalancers',
                          return_value=False).start()

    @contextlib.contextmanager
    def router(self, name='vpn-test-router', tenant_id=_uuid(),
               admin_state_up=True, **kwargs):
        request = {'router': {'tenant_id': tenant_id,
                              'name': name,
                              'admin_state_up': admin_state_up}}
        for arg in kwargs:
            request['router'][arg] = kwargs[arg]
        router = self.l3plugin.create_router(self.context, request)
        yield router

    def test_create_ipsec_site_connection(self):
        with mock.patch.object(self.service_plugin, 'get_ikepolicy',
                               return_value=FAKE_IKE_POLICY),\
            mock.patch.object(self.service_plugin, 'get_ipsecpolicy',
                              return_value=FAKE_IPSEC_POLICY),\
            mock.patch.object(self.service_plugin, '_get_vpnservice',
                              return_value=FAKE_VPNSERVICE),\
            mock.patch.object(self.service_plugin, 'get_vpnservices',
                              return_value=[FAKE_VPNSERVICE]),\
            mock.patch.object(self.plugin, 'get_router',
                              return_value=FAKE_ROUTER),\
            mock.patch.object(self.plugin, 'get_subnet',
                              return_value=FAKE_SUBNET),\
            mock.patch("vmware_nsx.db.db.add_nsx_vpn_connection_mapping"),\
            mock.patch.object(self.policy_vpn.ike_profile,
                              'create_or_overwrite') as create_ike,\
            mock.patch.object(self.policy_vpn.tunnel_profile,
                              'create_or_overwrite') as create_ipsec,\
            mock.patch.object(self.policy_vpn.dpd_profile,
                              'create_or_overwrite') as create_dpd,\
            mock.patch.object(self.policy_vpn.session,
                              'create_or_overwrite') as create_sesson:
            self.driver.create_ipsec_site_connection(self.context,
                                                     FAKE_IPSEC_CONNECTION)
            create_ike.assert_called_once()
            create_ipsec.assert_called_once()
            create_dpd.assert_called_once()
            create_sesson.assert_called_once()
            # TODO(asarfaty): make sure router adv also updated

    def test_update_ipsec_site_connection(self):
        with mock.patch.object(self.service_plugin, '_get_vpnservice',
                               return_value=FAKE_VPNSERVICE),\
            mock.patch.object(self.plugin, 'get_router',
                              return_value=FAKE_ROUTER),\
            mock.patch.object(self.plugin,
                              'update_router_firewall') as update_fw,\
            mock.patch.object(self.policy_vpn.session,
                              'update') as update_sesson,\
            mock.patch("vmware_nsx.db.db.get_nsx_vpn_connection_mapping"):
            self.driver.update_ipsec_site_connection(self.context,
                                                     FAKE_IPSEC_CONNECTION,
                                                     FAKE_NEW_CONNECTION)
            update_sesson.assert_called_once()
            update_fw.assert_called_once()

    def test_delete_ipsec_site_connection(self):
        with mock.patch.object(self.service_plugin, 'get_ikepolicy',
                               return_value=FAKE_IKE_POLICY),\
            mock.patch.object(self.service_plugin, 'get_ipsecpolicy',
                              return_value=FAKE_IPSEC_POLICY),\
            mock.patch.object(self.service_plugin, '_get_vpnservice',
                              return_value=FAKE_VPNSERVICE),\
            mock.patch.object(self.service_plugin, 'get_vpnservices',
                              return_value=[FAKE_VPNSERVICE]),\
            mock.patch.object(self.plugin, 'get_router',
                              return_value=FAKE_ROUTER),\
            mock.patch.object(self.plugin, 'get_subnet',
                              return_value=FAKE_SUBNET),\
            mock.patch("vmware_nsx.db.db.get_nsx_vpn_connection_mapping"),\
            mock.patch.object(self.policy_vpn.ike_profile,
                              'delete') as delete_ike,\
            mock.patch.object(self.policy_vpn.tunnel_profile,
                              'delete') as delete_ipsec,\
            mock.patch.object(self.policy_vpn.dpd_profile,
                              'delete') as delete_dpd,\
            mock.patch.object(self.policy_vpn.session,
                              'delete') as delete_sesson:
            self.driver.delete_ipsec_site_connection(self.context,
                                                     FAKE_IPSEC_CONNECTION)
            delete_ike.assert_called_once()
            delete_ipsec.assert_called_once()
            delete_dpd.assert_called_once()
            delete_sesson.assert_called_once()
            # TODO(asarfaty): make sure router adv rules also updated

    def test_create_vpn_service_legal(self):
        """Create a legal vpn service"""
        # create an external network with a subnet, and a router
        providernet_args = {extnet_apidef.EXTERNAL: True}
        router_db = namedtuple("Router", FAKE_ROUTER.keys())(
            *FAKE_ROUTER.values())
        tier0_uuid = 'tier-0'
        with self.network(name='ext-net',
                          providernet_args=providernet_args,
                          arg_list=(extnet_apidef.EXTERNAL, )) as ext_net,\
            self.subnet(ext_net, enable_dhcp=False),\
            mock.patch.object(self.plugin, '_get_tier0_uuid_by_router',
                              return_value=tier0_uuid),\
            self.router(external_gateway_info={'network_id':
                        ext_net['network']['id']}) as router,\
            self.subnet(cidr='1.1.0.0/24') as sub:
            # add an interface to the router
            self.l3plugin.add_router_interface(
                self.context,
                router['id'],
                {'subnet_id': sub['subnet']['id']})
            # create the service
            dummy_port = {'id': 'dummy_port',
                          'fixed_ips': [{'ip_address': '1.1.1.1'}]}
            tier0_rtr = {'ha_mode': 'ACTIVE_STANDBY'}
            with mock.patch.object(self.service_plugin, '_get_vpnservice',
                                   return_value=FAKE_VPNSERVICE),\
                mock.patch.object(self.policy_vpn.service,
                                  'create_or_overwrite') as create_service,\
                mock.patch.object(self.l3plugin, '_get_router',
                                  return_value=router_db),\
                mock.patch.object(self.plugin, 'get_router',
                                  return_value=FAKE_ROUTER),\
                mock.patch.object(self.plugin, 'get_ports',
                                  return_value=[dummy_port]),\
                mock.patch.object(self.plugin, 'delete_port') as delete_port,\
                mock.patch.object(self.plugin, 'service_router_has_services',
                                  return_value=True),\
                mock.patch.object(self.plugin.nsxpolicy.tier0, 'get',
                                  return_value=tier0_rtr):
                self.driver.create_vpnservice(self.context, FAKE_VPNSERVICE)
                create_service.assert_called_once()
                # Delete the service
                nsx_services = [{'logical_router_id': tier0_uuid,
                                 'id': 'xxx'}]
                with mock.patch.object(
                    self.policy_vpn.service, 'list',
                    return_value={'results': nsx_services}),\
                    mock.patch.object(self.service_plugin, 'get_vpnservices',
                                      return_value=[]),\
                    mock.patch.object(self.policy_vpn.service,
                                      'delete') as delete_service:
                    self.driver.delete_vpnservice(
                        self.context, FAKE_VPNSERVICE)
                    delete_service.assert_called_once()
                    delete_port.assert_called_once()

    def test_create_another_vpn_service(self):
        # make sure another backend service is not created
        providernet_args = {extnet_apidef.EXTERNAL: True}
        router_db = namedtuple("Router", FAKE_ROUTER.keys())(
            *FAKE_ROUTER.values())
        tier0_rtr_id = _uuid()
        with self.network(name='ext-net',
                          providernet_args=providernet_args,
                          arg_list=(extnet_apidef.EXTERNAL, )) as ext_net,\
            self.subnet(ext_net, enable_dhcp=False),\
            mock.patch.object(self.plugin, '_get_tier0_uuid_by_router',
                              return_value=tier0_rtr_id),\
            self.router(external_gateway_info={'network_id':
                        ext_net['network']['id']}) as router,\
            self.subnet(cidr='1.1.0.0/24') as sub:
            # add an interface to the router
            self.l3plugin.add_router_interface(
                self.context,
                router['id'],
                {'subnet_id': sub['subnet']['id']})
            # create the service
            dummy_port = {'id': 'dummy_port',
                          'fixed_ips': [{'ip_address': '1.1.1.1'}]}
            tier0_rtr = {'id': tier0_rtr_id,
                         'ha_mode': 'ACTIVE_STANDBY'}
            nsx_srv = {'logical_router_id': tier0_rtr_id,
                       'id': _uuid(),
                       'enabled': True}
            with mock.patch.object(self.service_plugin, '_get_vpnservice',
                                   return_value=FAKE_VPNSERVICE),\
                mock.patch.object(self.policy_vpn.service,
                                  'create_or_overwrite') as create_service,\
                mock.patch.object(
                    self.policy_vpn.service, 'list',
                    return_value={'results': [nsx_srv]}),\
                mock.patch.object(self.l3plugin, '_get_router',
                                  return_value=router_db),\
                mock.patch.object(self.plugin, 'get_router',
                                  return_value=FAKE_ROUTER),\
                mock.patch.object(self.plugin, 'get_ports',
                                  return_value=[dummy_port]),\
                mock.patch.object(self.plugin, 'delete_port'),\
                mock.patch.object(self.plugin, 'service_router_has_services',
                                  return_value=True),\
                mock.patch.object(self.plugin.nsxpolicy.tier0, 'get',
                                  return_value=tier0_rtr):
                self.driver.create_vpnservice(self.context, FAKE_VPNSERVICE)
                create_service.assert_called_once()

                # now delete both
                nsx_services = [{'logical_router_id': tier0_rtr_id,
                                 'id': 'xxx'}]
                with mock.patch.object(
                    self.policy_vpn.service, 'list',
                    return_value={'results': nsx_services}),\
                    mock.patch.object(self.policy_vpn.service,
                                      'delete') as delete_service:
                    self.driver.delete_vpnservice(
                        self.context, FAKE_VPNSERVICE)
                    delete_service.assert_not_called()

                with mock.patch.object(
                    self.policy_vpn.service, 'list',
                    return_value={'results': nsx_services}),\
                    mock.patch.object(self.service_plugin, 'get_vpnservices',
                                      return_value=[]),\
                    mock.patch.object(self.policy_vpn.service,
                                      'delete') as delete_service:
                    self.driver.delete_vpnservice(
                        self.context, FAKE_VPNSERVICE)
                    delete_service.assert_called_once()



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.db import db as nsx_db
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_common
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_v3.implementation import lb_utils
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)


class EdgeListenerManagerFromDict(base_mgr.Nsxv3LoadbalancerBaseManager):
    def _get_virtual_server_kwargs(self, context, listener, vs_name, tags,
                                   app_profile_id, certificate=None):
        # If loadbalancer vip_port already has floating ip, use floating
        # IP as the virtual server VIP address. Else, use the loadbalancer
        # vip_address directly on virtual server.
        filters = {'port_id': [listener['loadbalancer']['vip_port_id']]}
        floating_ips = self.core_plugin.get_floatingips(context,
                                                        filters=filters)
        if floating_ips:
            lb_vip_address = floating_ips[0]['floating_ip_address']
        else:
            lb_vip_address = listener['loadbalancer']['vip_address']
        kwargs = {'enabled': listener['admin_state_up'],
                  'ip_address': lb_vip_address,
                  'port': listener['protocol_port'],
                  'application_profile_id': app_profile_id,
                  'description': listener.get('description')}
        if vs_name:
            kwargs['display_name'] = vs_name
        if tags:
            kwargs['tags'] = tags
        if listener['connection_limit'] != -1:
            kwargs['max_concurrent_connections'] = \
                listener['connection_limit']
        if 'default_pool_id' in listener:
            if listener['default_pool_id']:
                pool_binding = nsx_db.get_nsx_lbaas_pool_binding(
                    context.session, listener['loadbalancer']['id'],
                    listener['default_pool_id'])
                if pool_binding:
                    kwargs['pool_id'] = pool_binding.get('lb_pool_id')
            else:
                # Remove the default pool
                kwargs['pool_id'] = None
                kwargs['persistence_profile_id'] = ''

        ssl_profile_binding = self._get_ssl_profile_binding(
            tags, certificate=certificate)
        if (listener['protocol'] == lb_const.LB_PROTOCOL_TERMINATED_HTTPS and
            ssl_profile_binding):
            kwargs.update(ssl_profile_binding)
        return kwargs

    def _get_ssl_profile_binding(self, tags, certificate=None):
        tm_client = self.core_plugin.nsxlib.trust_management
        if certificate:
            # First check if NSX already has certificate with same pem.
            # If so, use that certificate for ssl binding. Otherwise,
            # create a new certificate on NSX.
            cert_ids = tm_client.find_cert_with_pem(
                certificate.get_certificate())
            if cert_ids:
                nsx_cert_id = cert_ids[0]
            else:
                nsx_cert_id = tm_client.create_cert(
                    certificate.get_certificate(),
                    private_key=certificate.get_private_key(),
                    passphrase=certificate.get_private_key_passphrase(),
                    tags=tags)
            return {
                'client_ssl_profile_binding': {
                    'ssl_profile_id': self.core_plugin.client_ssl_profile,
                    'default_certificate_id': nsx_cert_id
                }
            }

    def _get_listener_tags(self, context, listener):
        tags = lb_utils.get_tags(self.core_plugin, listener['id'],
                                 lb_const.LB_LISTENER_TYPE,
                                 listener['tenant_id'],
                                 context.project_name)
        tags.append({
            'scope': lb_const.LB_LB_NAME,
            'tag': listener['loadbalancer']['name'][:utils.MAX_TAG_LEN]})
        tags.append({
            'scope': lb_const.LB_LB_TYPE,
            'tag': listener['loadbalancer_id']})
        return tags

    def _validate_default_pool(self, context, listener, vs_id, completor,
                               old_listener=None):
        if listener.get('default_pool_id'):
            pool_binding = nsx_db.get_nsx_lbaas_pool_binding(
                context.session, listener['loadbalancer']['id'],
                listener['default_pool_id'])
            if (pool_binding and pool_binding['lb_vs_id'] and
                (vs_id is None or pool_binding['lb_vs_id'] != vs_id)):
                completor(success=False)
                msg = (_('Default pool %s is already used by another '
                         'listener') % listener['default_pool_id'])
                raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)

            lb_common.validate_session_persistence(
                listener.get('default_pool'), listener, completor)

    def _update_default_pool_and_binding(self, context, listener, vs_data,
                                         completor, old_listener=None):
        vs_client = self.core_plugin.nsxlib.load_balancer.virtual_server
        if listener.get('default_pool_id'):
            vs_id = vs_data['id']
            lb_id = (listener.get('loadbalancer_id') or
                     listener.get('loadbalancer', {}).get('id'))
            pool_id = listener['default_pool_id']
            pool = listener['default_pool']
            old_pool = None
            if old_listener:
                old_pool = old_listener.get('default_pool')
            try:
                switch_type = lb_common.session_persistence_type_changed(
                    pool, old_pool)
                (persistence_profile_id,
                 post_process_func) = lb_utils.setup_session_persistence(
                    self.core_plugin.nsxlib,
                    pool,
                    lb_utils.get_pool_tags(context, self.core_plugin, pool),
                    switch_type, listener, vs_data)
            except nsxlib_exc.ManagerError:
                with excutils.save_and_reraise_exception():
                    completor(success=False)
                    LOG.error("Failed to configure session persistence "
                              "profile for listener %s", listener['id'])
            try:
                # Update persistence profile and pool on virtual server
                vs_client.update(
                    vs_id,
                    persistence_profile_id=persistence_profile_id)
                LOG.debug("Updated NSX virtual server %(vs_id)s with "
                          "persistence profile %(prof)s",
                          {'vs_id': vs_id,
                           'prof': persistence_profile_id})
                if post_process_func:
                    post_process_func()
            except nsxlib_exc.ManagerError:
                with excutils.save_and_reraise_exception():
                    completor(success=False)
                    LOG.error("Failed to attach persistence profile %s to "
                              "virtual server %s",
                              persistence_profile_id, vs_id)
            # Update the DB binding of the default pool
            nsx_db.update_nsx_lbaas_pool_binding(
                context.session, lb_id, pool_id, vs_id)

    def _remove_default_pool_binding(self, context, listener):
        if not listener.get('default_pool_id'):
            return

        # Remove the current default pool from the DB bindings
        lb_id = (listener.get('loadbalancer_id') or
                 listener.get('loadbalancer', {}).get('id'))
        pool_id = listener['default_pool_id']
        pool_binding = nsx_db.get_nsx_lbaas_pool_binding(
            context.session, lb_id, pool_id)
        if pool_binding:
            nsx_db.update_nsx_lbaas_pool_binding(
                context.session, lb_id, pool_id, None)

    def create(self, context, listener, completor,
               certificate=None):
        lb_id = listener['loadbalancer_id']
        nsxlib_lb = self.core_plugin.nsxlib.load_balancer
        app_client = nsxlib_lb.application_profile
        vs_client = nsxlib_lb.virtual_server
        service_client = nsxlib_lb.service
        vs_name = utils.get_name_and_uuid(listener['name'] or 'listener',
                                          listener['id'])
        tags = self._get_listener_tags(context, listener)

        if (listener['protocol'] == lb_const.LB_PROTOCOL_HTTP or
                listener['protocol'] == lb_const.LB_PROTOCOL_TERMINATED_HTTPS):
            profile_type = lb_const.LB_HTTP_PROFILE
        elif (listener['protocol'] == lb_const.LB_PROTOCOL_TCP or
              listener['protocol'] == lb_const.LB_PROTOCOL_HTTPS):
            profile_type = lb_const.LB_TCP_PROFILE
        else:
            completor(success=False)
            msg = (_('Cannot create listener %(listener)s with '
                     'protocol %(protocol)s') %
                   {'listener': listener['id'],
                    'protocol': listener['protocol']})
            raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)
        # Validate default pool
        self._validate_default_pool(context, listener, None, completor)
        try:
            app_profile = app_client.create(
                display_name=vs_name, resource_type=profile_type, tags=tags)
            app_profile_id = app_profile['id']
            kwargs = self._get_virtual_server_kwargs(
                context, listener, vs_name, tags, app_profile_id, certificate)
            virtual_server = vs_client.create(**kwargs)
        except nsxlib_exc.ManagerError:
            completor(success=False)
            msg = _('Failed to create virtual server at NSX backend')
            raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)

        # If there is already lb:lb_service binding, add the virtual
        # server to the lb service
        binding = nsx_db.get_nsx_lbaas_loadbalancer_binding(
            context.session, lb_id)
        if not binding:
            completor(success=False)
            msg = _('Failed to get loadbalancer %s binding') % lb_id
            raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)

        lb_service_id = binding['lb_service_id']
        try:
            service_client.add_virtual_server(lb_service_id,
                                              virtual_server['id'])
        except nsxlib_exc.ManagerError:
            completor(success=False)
            msg = _('Failed to add virtual server to lb service '
                    'at NSX backend')
            # delete the backend virtual server
            vs_client.delete(virtual_server['id'])
            raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)

        nsx_db.add_nsx_lbaas_listener_binding(
            context.session, lb_id, listener['id'], app_profile_id,
            virtual_server['id'])
        self._update_default_pool_and_binding(
            context, listener, virtual_server, completor)
        completor(success=True)

    def update(self, context, old_listener, new_listener, completor,
               certificate=None):
        nsxlib_lb = self.core_plugin.nsxlib.load_balancer
        vs_client = nsxlib_lb.virtual_server
        app_client = nsxlib_lb.application_profile
        vs_name = None
        tags = None
        if new_listener['name'] != old_listener['name']:
            vs_name = utils.get_name_and_uuid(
                new_listener['name'] or 'listener',
                new_listener['id'])
            tags = self._get_listener_tags(context, new_listener)

        binding = nsx_db.get_nsx_lbaas_listener_binding(
            context.session, old_listener['loadbalancer_id'],
            old_listener['id'])
        if not binding:
            msg = (_('Cannot find listener %(listener)s binding on NSX '
                     'backend'), {'listener': old_listener['id']})
            raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)

        # Validate default pool
        self._validate_default_pool(
            context, new_listener, binding['lb_vs_id'], completor,
            old_listener=old_listener)

        try:
            vs_id = binding['lb_vs_id']
            app_profile_id = binding['app_profile_id']
            updated_kwargs = self._get_virtual_server_kwargs(
                context, new_listener, vs_name, tags, app_profile_id,
                certificate)
            vs_data = vs_client.update(vs_id, **updated_kwargs)
            if vs_name:
                app_client.update(app_profile_id, display_name=vs_name,
                                  tags=tags)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error('Failed to update listener %(listener)s with '
                          'error %(error)s',
                          {'listener': old_listener['id'], 'error': e})
        # Update default pool and session persistence (do this even if the
        # default pool did not change, as there might have been an error the
        # last time)
        self._remove_default_pool_binding(context, old_listener)
        self._update_default_pool_and_binding(context, new_listener,
                                              vs_data, completor,
                                              old_listener)

        completor(success=True)

    def delete(self, context, listener, completor):
        lb_id = listener['loadbalancer_id']
        nsxlib_lb = self.core_plugin.nsxlib.load_balancer
        service_client = nsxlib_lb.service
        vs_client = nsxlib_lb.virtual_server
        app_client = nsxlib_lb.application_profile

        binding = nsx_db.get_nsx_lbaas_listener_binding(
            context.session, lb_id, listener['id'])
        if binding:
            vs_id = binding['lb_vs_id']
            app_profile_id = binding['app_profile_id']
            lb_binding = nsx_db.get_nsx_lbaas_loadbalancer_binding(
                context.session, lb_id)
            if not lb_binding:
                completor(success=False)
                msg = (_('Failed to delete virtual server: %(listener)s: '
                         'loadbalancer %(lb)s mapping was not found') %
                       {'listener': listener['id'], 'lb': lb_id})
                raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)
            try:
                lbs_id = lb_binding.get('lb_service_id')
                lb_service = service_client.get(lbs_id)
                vs_list = lb_service.get('virtual_server_ids')
                if vs_list and vs_id in vs_list:
                    service_client.remove_virtual_server(lbs_id, vs_id)
            except nsxlib_exc.ManagerError:
                completor(success=False)
                msg = (_('Failed to remove virtual server: %(listener)s '
                         'from lb service %(lbs)s') %
                       {'listener': listener['id'], 'lbs': lbs_id})
                raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)
            try:
                persist_profile_id = None
                if listener.get('default_pool_id'):
                    vs_data = vs_client.update(vs_id, pool_id='')
                    persist_profile_id = vs_data.get('persistence_profile_id')
                    # Update pool binding to disassociate virtual server
                    self._remove_default_pool_binding(context, listener)
                vs_client.delete(vs_id)
                # Also delete the old session persistence profile
                if persist_profile_id:
                    lb_utils.delete_persistence_profile(
                        self.core_plugin.nsxlib, persist_profile_id)
            except nsx_exc.NsxResourceNotFound:
                msg = (_("virtual server not found on nsx: %(vs)s") %
                       {'vs': vs_id})
                raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)
            except nsxlib_exc.ManagerError:
                completor(success=False)
                msg = (_('Failed to delete virtual server: %(listener)s') %
                       {'listener': listener['id']})
                raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)
            try:
                app_client.delete(app_profile_id)
            except nsx_exc.NsxResourceNotFound:
                LOG.error("application profile not found on nsx: %s",
                          app_profile_id)
            except nsxlib_exc.ManagerError as e:
                # This probably means that the application profile is being
                # used by a listener outside of openstack
                LOG.error("Failed to delete application profile %s from the "
                          "NSX: %s", app_profile_id, e)

            # Delete imported NSX cert if there is any
            cert_tags = [{'scope': lb_const.LB_LISTENER_TYPE,
                          'tag': listener['id']}]
            results = self.core_plugin.nsxlib.search_by_tags(
                tags=cert_tags)
            # Only delete object related to certificate used by listener
            for res_obj in results['results']:
                res_type = res_obj.get('resource_type')
                if res_type in lb_const.LB_CERT_RESOURCE_TYPE:
                    tm_client = self.core_plugin.nsxlib.trust_management
                    try:
                        tm_client.delete_cert(res_obj['id'])
                    except nsxlib_exc.ManagerError:
                        LOG.error("Exception thrown when trying to delete "
                                  "certificate: %(cert)s",
                                  {'cert': res_obj['id']})

            nsx_db.delete_nsx_lbaas_listener_binding(
                context.session, lb_id, listener['id'])

        completor(success=True)

    def delete_cascade(self, context, listener, completor):
        self.delete(context, listener, completor)


def stats_getter(context, core_plugin, ignore_list=None):
    """Update Octavia statistics for each listener (virtual server)"""
    stat_list = []
    lb_service_client = core_plugin.nsxlib.load_balancer.service
    # Go over all the loadbalancers & services
    lb_bindings = nsx_db.get_nsx_lbaas_loadbalancer_bindings(
        context.session)
    for lb_binding in lb_bindings:
        if ignore_list and lb_binding['loadbalancer_id'] in ignore_list:
            continue

        lb_service_id = lb_binding.get('lb_service_id')
        try:
            # get the NSX statistics for this LB service
            rsp = lb_service_client.get_stats(lb_service_id)
            if rsp and 'virtual_servers' in rsp:
                # Go over each virtual server in the response
                for vs in rsp['virtual_servers']:
                    # look up the virtual server in the DB
                    vs_bind = nsx_db.get_nsx_lbaas_listener_binding_by_vs_id(
                        context.session, vs['virtual_server_id'])
                    if vs_bind and 'statistics' in vs:
                        vs_stats = vs['statistics']
                        stats = copy.copy(lb_const.LB_EMPTY_STATS)
                        stats['id'] = vs_bind.listener_id
                        stats['request_errors'] = 0  # currently unsupported
                        for stat in lb_const.LB_STATS_MAP:
                            lb_stat = lb_const.LB_STATS_MAP[stat]
                            stats[stat] += vs_stats[lb_stat]
                        stat_list.append(stats)

        except nsxlib_exc.ManagerError:
            pass

    return stat_list



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.db import db as nsx_db
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_v3.implementation import lb_utils
from vmware_nsx.services.lbaas.octavia import constants as oct_const
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)


class EdgeLoadBalancerManagerFromDict(base_mgr.Nsxv3LoadbalancerBaseManager):

    def create(self, context, lb, completor):
        if not lb_utils.validate_lb_subnet(context, self.core_plugin,
                                           lb['vip_subnet_id']):
            completor(success=False)
            msg = (_('Cannot create lb on subnet %(sub)s for '
                     'loadbalancer %(lb)s. The subnet needs to connect a '
                     'router which is already set gateway.') %
                   {'sub': lb['vip_subnet_id'], 'lb': lb['id']})
            raise n_exc.BadRequest(resource='lbaas-subnet', msg=msg)

        service_client = self.core_plugin.nsxlib.load_balancer.service
        nsx_router_id = None
        lb_service = None
        nsx_router_id = lb_utils.NO_ROUTER_ID
        router_id = lb_utils.get_router_from_network(
            context, self.core_plugin, lb['vip_subnet_id'])
        if router_id:
            nsx_router_id = nsx_db.get_nsx_router_id(context.session,
                                                     router_id)
            lb_service = service_client.get_router_lb_service(nsx_router_id)
        if not lb_service:
            lb_size = lb_utils.get_lb_flavor_size(
                self.flavor_plugin, context, lb.get('flavor_id'))
            if router_id:
                # Make sure the NSX service router exists
                if not self.core_plugin.service_router_has_services(
                        context, router_id):
                    self.core_plugin.create_service_router(context, router_id)
                lb_service = self._create_lb_service(
                    context, service_client, lb['tenant_id'],
                    router_id, nsx_router_id, lb['id'], lb_size)
            else:
                lb_service = self._create_lb_service_without_router(
                    context, service_client, lb['tenant_id'],
                    lb, lb_size)
            if not lb_service:
                completor(success=False)
                msg = (_('Failed to create lb service for loadbalancer '
                         '%s') % lb['id'])
                raise nsx_exc.NsxPluginException(err_msg=msg)

        nsx_db.add_nsx_lbaas_loadbalancer_binding(
            context.session, lb['id'], lb_service['id'],
            nsx_router_id, lb['vip_address'])

        # Make sure the vip port is marked with a device owner
        port = self.core_plugin.get_port(
            context.elevated(), lb['vip_port_id'])
        if not port.get('device_owner'):
            self.core_plugin.update_port(
                context.elevated(), lb['vip_port_id'],
                {'port': {'device_id': oct_const.DEVICE_ID_PREFIX + lb['id'],
                          'device_owner': lb_const.VMWARE_LB_VIP_OWNER}})

        completor(success=True)

    def _create_lb_service(self, context, service_client, tenant_id,
                           router_id, nsx_router_id, lb_id, lb_size):
        """Create NSX LB service for a specific neutron router"""
        router = self.core_plugin.get_router(context, router_id)
        if not router.get('external_gateway_info'):
            msg = (_('Tenant router %(router)s does not connect to '
                     'external gateway') % {'router': router['id']})
            raise n_exc.BadRequest(resource='lbaas-lbservice-create',
                                   msg=msg)
        lb_name = utils.get_name_and_uuid(router['name'] or 'router',
                                          router_id)
        tags = lb_utils.get_tags(self.core_plugin, router_id,
                                 lb_const.LR_ROUTER_TYPE,
                                 tenant_id, context.project_name)
        attachment = {'target_id': nsx_router_id,
                      'target_type': 'LogicalRouter'}
        try:
            lb_service = service_client.create(display_name=lb_name,
                                               tags=tags,
                                               attachment=attachment,
                                               size=lb_size)
        except nsxlib_exc.ManagerError as e:
            # If it failed, it is probably because the service was already
            # created by another loadbalancer simultaneously
            lb_service = service_client.get_router_lb_service(nsx_router_id)
            if lb_service:
                return lb_service
            LOG.error("Failed to create LB service: %s", e)
            return

        # Add rule to advertise external vips
        lb_utils.update_router_lb_vip_advertisement(
            context, self.core_plugin, router, nsx_router_id)

        return lb_service

    def _create_lb_service_without_router(self, context, service_client,
                                          tenant_id, lb, lb_size):
        """Create NSX LB service for an external VIP
        This service will not be attached to a router yet, and it will be
        updated once the first member is created.
        """
        lb_id = lb['id']
        lb_name = utils.get_name_and_uuid(lb['name'] or 'loadbalancer',
                                          lb_id)
        tags = lb_utils.get_tags(self.core_plugin, '',
                                 lb_const.LR_ROUTER_TYPE,
                                 tenant_id, context.project_name)
        try:
            lb_service = service_client.create(display_name=lb_name,
                                               tags=tags,
                                               size=lb_size)
        except nsxlib_exc.ManagerError as e:
            LOG.error("Failed to create LB service: %s", e)
            return

        return lb_service

    def update(self, context, old_lb, new_lb, completor):
        vs_client = self.core_plugin.nsxlib.load_balancer.virtual_server
        app_client = self.core_plugin.nsxlib.load_balancer.application_profile
        if new_lb['name'] != old_lb['name']:
            for listener in new_lb['listeners']:
                binding = nsx_db.get_nsx_lbaas_listener_binding(
                    context.session, new_lb['id'], listener['id'])
                if binding:
                    vs_id = binding['lb_vs_id']
                    app_profile_id = binding['app_profile_id']
                    new_lb_name = new_lb['name'][:utils.MAX_TAG_LEN]
                    try:
                        # Update tag on virtual server with new lb name
                        vs = vs_client.get(vs_id)
                        updated_tags = utils.update_v3_tags(
                            vs['tags'], [{'scope': lb_const.LB_LB_NAME,
                                          'tag': new_lb_name}])
                        vs_client.update(vs_id, tags=updated_tags)
                        # Update tag on application profile with new lb name
                        app_profile = app_client.get(app_profile_id)
                        app_client.update(
                            app_profile_id, tags=updated_tags,
                            resource_type=app_profile['resource_type'])

                    except nsxlib_exc.ManagerError:
                        with excutils.save_and_reraise_exception():
                            completor(success=False)
                            LOG.error('Failed to update tag %(tag)s for lb '
                                      '%(lb)s', {'tag': updated_tags,
                                                 'lb': new_lb['name']})

        completor(success=True)

    def delete(self, context, lb, completor):
        service_client = self.core_plugin.nsxlib.load_balancer.service
        router_client = self.core_plugin.nsxlib.logical_router
        lb_binding = nsx_db.get_nsx_lbaas_loadbalancer_binding(
            context.session, lb['id'])
        if lb_binding:
            lb_service_id = lb_binding['lb_service_id']
            nsx_router_id = lb_binding['lb_router_id']
            try:
                lb_service = service_client.get(lb_service_id)
            except nsxlib_exc.ManagerError:
                LOG.warning("LB service %(lbs)s is not found",
                            {'lbs': lb_service_id})
            else:
                vs_list = lb_service.get('virtual_server_ids')
                if not vs_list:
                    try:
                        service_client.delete(lb_service_id)
                        # If there is no lb service attached to the router,
                        # delete the router advertise_lb_vip rule.
                        if nsx_router_id != lb_utils.NO_ROUTER_ID:
                            router_client.update_advertisement_rules(
                                nsx_router_id, [],
                                name_prefix=lb_utils.ADV_RULE_NAME)
                    except nsxlib_exc.ManagerError:
                        completor(success=False)
                        msg = (_('Failed to delete lb service %(lbs)s from nsx'
                                 ) % {'lbs': lb_service_id})
                        raise n_exc.BadRequest(resource='lbaas-lb', msg=msg)
            nsx_db.delete_nsx_lbaas_loadbalancer_binding(
                context.session, lb['id'])
            if nsx_router_id != lb_utils.NO_ROUTER_ID:
                router_id = nsx_db.get_neutron_from_nsx_router_id(
                    context.session, nsx_router_id)
                # Service router is needed only when the LB exist, and
                # no other services are using it.
                if not self.core_plugin.service_router_has_services(
                        context,
                        router_id):
                    self.core_plugin.delete_service_router(context,
                                                           router_id)
        # Make sure the vip port is not marked with a vmware device owner
        try:
            port = self.core_plugin.get_port(
                context.elevated(), lb['vip_port_id'])
            if port.get('device_owner') == lb_const.VMWARE_LB_VIP_OWNER:
                self.core_plugin.update_port(
                    context.elevated(), lb['vip_port_id'],
                    {'port': {'device_id': '',
                              'device_owner': ''}})
        except n_exc.PortNotFound:
            # Only log the error and continue anyway
            LOG.warning("VIP port %s not found while deleting loadbalancer %s",
                        lb['vip_port_id'], lb['id'])
        except Exception as e:
            # Just log the error as all other resources were deleted
            LOG.error("Failed to update neutron port %s devices upon "
                      "loadbalancer deletion: %s", lb['vip_port_id'], e)

        completor(success=True)

    def delete_cascade(self, context, lb, completor):
        """Delete all backend and DB resources of this loadbalancer"""
        self.delete(context, lb, completor)

    def refresh(self, context, lb):
        # TODO(tongl): implement
        pass

    def _nsx_status_to_lb_status(self, nsx_status):
        if not nsx_status:
            # default fallback
            return lb_const.ONLINE

        # Statuses that are considered ONLINE:
        if nsx_status.upper() in ['UP', 'UNKNOWN', 'PARTIALLY_UP',
                                  'NO_STANDBY']:
            return lb_const.ONLINE
        # Statuses that are considered OFFLINE:
        if nsx_status.upper() in ['PRIMARY_DOWN', 'DETACHED', 'DOWN', 'ERROR']:
            return lb_const.OFFLINE
        if nsx_status.upper() == 'DISABLED':
            return lb_const.DISABLED

        # default fallback
        LOG.debug("NSX LB status %s - interpreted as ONLINE", nsx_status)
        return lb_const.ONLINE

    def get_lb_pool_members_statuses(self, nsx_pool_id, members_statuses):
        # Combine the NSX pool members data and the NSX statuses to provide
        # member statuses list
        # Get the member id from the suffix of the member in the NSX pool list
        # and find the matching ip+port member in the statuses list
        # get the members list from the NSX
        nsx_pool = self.core_plugin.nsxlib.load_balancer.pool.get(nsx_pool_id)
        if not nsx_pool or not nsx_pool.get('members'):
            return []
        # create a map of existing members: ip+port -> lbaas ID (which is the
        # suffix of the member name)
        members_map = {}
        for member in nsx_pool['members']:
            ip = member['ip_address']
            port = member['port']
            if ip not in members_map:
                members_map[ip] = {}
            members_map[ip][port] = member['display_name'][-36:]
        # go over the statuses map, and match the member ip_port, to the ID
        # in the map
        statuses = []
        for member in members_statuses:
            ip = member['ip_address']
            port = member['port']
            if ip in members_map and port in members_map[ip]:
                member_id = members_map[ip][port]
                member_status = self._nsx_status_to_lb_status(member['status'])
                statuses.append({'id': member_id, 'status': member_status})
        return statuses

    def get_operating_status(self, context, id, with_members=False):
        """Return a map of the operating status of all connected LB objects """
        service_client = self.core_plugin.nsxlib.load_balancer.service
        lb_binding = nsx_db.get_nsx_lbaas_loadbalancer_binding(
            context.session, id)
        if not lb_binding:
            LOG.warning("Failed to get loadbalancer %s operating status. "
                        "Mapping was not found", id)
            return {}

        lb_service_id = lb_binding['lb_service_id']
        try:
            service_status = service_client.get_status(lb_service_id)
            if not isinstance(service_status, dict):
                service_status = {}

            vs_statuses = service_client.get_virtual_servers_status(
                lb_service_id)
            if not isinstance(vs_statuses, dict):
                vs_statuses = {}
        except nsxlib_exc.ManagerError:
            LOG.warning("LB service %(lbs)s is not found",
                        {'lbs': lb_service_id})
            return {}

        # get the loadbalancer status from the LB service
        lb_status = self._nsx_status_to_lb_status(
            service_status.get('service_status'))
        statuses = {lb_const.LOADBALANCERS: [{'id': id, 'status': lb_status}],
                    lb_const.LISTENERS: [],
                    lb_const.POOLS: [],
                    lb_const.MEMBERS: []}

        # Add the listeners statuses from the virtual servers statuses
        for vs in vs_statuses.get('results', []):
            vs_status = self._nsx_status_to_lb_status(vs.get('status'))
            vs_id = vs.get('virtual_server_id')
            list_binding = nsx_db.get_nsx_lbaas_listener_binding_by_lb_and_vs(
                context.session, id, vs_id)
            if list_binding:
                listener_id = list_binding['listener_id']
                statuses[lb_const.LISTENERS].append(
                    {'id': listener_id, 'status': vs_status})

        # Add the pools statuses from the LB service status
        for pool in service_status.get('pools', []):
            nsx_pool_id = pool.get('pool_id')
            pool_status = self._nsx_status_to_lb_status(pool.get('status'))
            pool_binding = nsx_db.get_nsx_lbaas_pool_binding_by_lb_pool(
                context.session, id, nsx_pool_id)
            if pool_binding:
                pool_id = pool_binding['pool_id']
                statuses[lb_const.POOLS].append(
                    {'id': pool_id, 'status': pool_status})
                # Add the pools members
                if with_members and pool.get('members'):
                    statuses[lb_const.MEMBERS].extend(
                        self.get_lb_pool_members_statuses(
                            nsx_pool_id, pool['members']))

        return statuses

    def stats(self, context, lb):
        # Since multiple LBaaS loadbalancer can share the same LB service,
        # get the corresponding virtual servers' stats instead of LB service.
        stats = {'active_connections': 0,
                 'bytes_in': 0,
                 'bytes_out': 0,
                 'total_connections': 0}

        service_client = self.core_plugin.nsxlib.load_balancer.service
        lb_binding = nsx_db.get_nsx_lbaas_loadbalancer_binding(
            context.session, lb['id'])
        vs_list = self._get_lb_virtual_servers(context, lb)
        if lb_binding:
            lb_service_id = lb_binding.get('lb_service_id')
            try:
                rsp = service_client.get_stats(lb_service_id)
                if rsp:
                    for vs in rsp.get('virtual_servers', []):
                        # Skip the virtual server that doesn't belong
                        # to this loadbalancer
                        if vs['virtual_server_id'] not in vs_list:
                            continue
                        vs_stats = vs.get('statistics', {})
                        for stat in lb_const.LB_STATS_MAP:
                            lb_stat = lb_const.LB_STATS_MAP[stat]
                            stats[stat] += vs_stats.get(lb_stat, 0)

            except nsxlib_exc.ManagerError:
                msg = _('Failed to retrieve stats from LB service '
                        'for loadbalancer %(lb)s') % {'lb': lb['id']}
                raise n_exc.BadRequest(resource='lbaas-lb', msg=msg)
        return stats

    def _get_lb_virtual_servers(self, context, lb):
        # Get all virtual servers that belong to this loadbalancer
        vs_list = []
        for listener in lb['listeners']:
            vs_binding = nsx_db.get_nsx_lbaas_listener_binding(
                context.session, lb['id'], listener['id'])
            if vs_binding:
                vs_list.append(vs_binding.get('lb_vs_id'))
        return vs_list



# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

import netaddr
from oslo_log import log as logging

from neutron_lib.exceptions import firewall_v2 as exceptions

from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.services.fwaas.common import fwaas_callbacks_v2 as \
    com_callbacks
from vmware_nsx.services.fwaas.common import v3_utils
from vmware_nsxlib.v3 import exceptions as nsx_lib_exc
from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3.policy import constants as policy_constants
from vmware_nsxlib.v3 import utils as nsxlib_utils

LOG = logging.getLogger(__name__)
GATEWAY_POLICY_NAME = 'Tier1 %s gateway policy'
DEFAULT_RULE_NAME = 'Default LR Layer3 Rule'
DEFAULT_RULE_ID = 'default_rule'
RULE_NAME_PREFIX = 'Fwaas-'
ROUTER_FW_TAG = 'os-router-firewall'


class NsxpFwaasCallbacksV2(com_callbacks.NsxCommonv3FwaasCallbacksV2):
    """NSX-P RPC callbacks for Firewall As A Service V2."""

    def __init__(self, with_rpc):
        super(NsxpFwaasCallbacksV2, self).__init__(with_rpc)
        self.internal_driver = None
        if self.fwaas_enabled:
            self.internal_driver = self.fwaas_driver

    @property
    def plugin_type(self):
        return projectpluginmap.NsxPlugins.NSX_P

    @property
    def nsxpolicy(self):
        return self.core_plugin.nsxpolicy

    def _get_default_backend_rule(self, router_id):
        """Return the default allow-all rule entry

        This rule entry will be added to the end of the rules list
        """
        return self.nsxpolicy.gateway_policy.build_entry(
            DEFAULT_RULE_NAME,
            policy_constants.DEFAULT_DOMAIN, router_id,
            self._get_random_rule_id(DEFAULT_RULE_ID),
            description=DEFAULT_RULE_NAME,
            sequence_number=None,
            action=nsx_constants.FW_ACTION_ALLOW,
            scope=[self.nsxpolicy.tier1.get_path(router_id)],
            source_groups=None, dest_groups=None,
            direction=nsx_constants.IN_OUT)

    def _translate_service(self, project_id, router_id, rule):
        """Return the NSX Policy service id matching the FW rule service.

        L4 protocol service will be created per router-id & rule-id
        and the service id will reflect both, as will as the L4 protocol.
        This will allow the cleanup of the service by tags when the router is
        detached.
        """
        ip_version = rule.get('ip_version', 4)
        if rule.get('protocol'):
            tags = self.nsxpolicy.build_v3_tags_payload(
                rule, resource_type='os-neutron-fwrule-id',
                project_name=project_id)
            tags = nsxlib_utils.add_v3_tag(tags, ROUTER_FW_TAG, router_id)
            l4_protocol = v3_utils.translate_fw_rule_protocol(
                rule.get('protocol'))
            srv_name = 'FW_rule_%s_%s_service' % (rule['id'], rule['protocol'])
            description = '%s service for FW rule %s of Tier1 %s' % (
                rule['protocol'], rule['id'], router_id)
            if l4_protocol in [nsx_constants.TCP, nsx_constants.UDP]:
                if rule.get('destination_port') is None:
                    destination_ports = []
                else:
                    destination_ports = v3_utils.translate_fw_rule_ports(
                        rule['destination_port'])

                if rule.get('source_port') is None:
                    source_ports = []
                else:
                    source_ports = v3_utils.translate_fw_rule_ports(
                        rule['source_port'])

                srv_id = self.nsxpolicy.service.create_or_overwrite(
                    srv_name,
                    description=description,
                    protocol=l4_protocol,
                    dest_ports=destination_ports,
                    source_ports=source_ports,
                    tags=tags)
            elif l4_protocol == nsx_constants.ICMPV4:
                #TODO(asarfaty): Can use predefined service for ICMP
                srv_id = self.nsxpolicy.icmp_service.create_or_overwrite(
                    srv_name,
                    version=ip_version,
                    tags=tags)
            return srv_id

    def _get_random_rule_id(self, rule_id):
        """Return a rule ID with random suffix to be used on the NSX
        Random sequence needs to be added to rule IDs, so that PUT command
        will replace all existing rules.
        Keeping the same rule id will require updating the rule revision as
        well.
        """
        #TODO(asarfaty): add support for self created id in build_entry and
        # remove this method
        return '%s-%s' % (rule_id, str(random.randint(1, 10000000)))

    def _get_rule_ips_group_id(self, rule_id, direction):
        return '%s-%s' % (direction, rule_id)

    def _is_empty_cidr(self, cidr, fwaas_rule_id):
        net = netaddr.IPNetwork(cidr)
        if ((net.version == 4 and cidr.startswith('0.0.0.0')) or
            (net.version == 6 and str(net.ip) == "::")):
            LOG.warning("Unsupported FWaaS cidr %(cidr)s for rule %(id)s",
                        {'cidr': cidr, 'id': fwaas_rule_id})
            return True

    def _validate_cidr(self, cidr, fwaas_rule_id):
        error_msg = (_("Illegal FWaaS cidr %(cidr)s for rule %(id)s") %
                     {'cidr': cidr, 'id': fwaas_rule_id})
        # Validate that this is a legal & supported ipv4 / ipv6 cidr
        net = netaddr.IPNetwork(cidr)
        if net.version == 4:
            if net.prefixlen == 0:
                LOG.error(error_msg)
                raise self.driver_exception(driver=self.driver_name)
        elif net.version == 6:
            if net.prefixlen == 0:
                LOG.error(error_msg)
                raise self.driver_exception(driver=self.driver_name)
        else:
            LOG.error(error_msg)
            raise self.driver_exception(driver=self.driver_name)

    def _get_rule_cidr_group(self, project_id, router_id, rule, is_source,
                             is_ingress):
        field = 'source_ip_address' if is_source else 'destination_ip_address'
        direction_text = 'source' if is_source else 'destination'
        if (rule.get(field) and
            not self._is_empty_cidr(rule[field], rule['id'])):
            # Create a group for ips
            group_ips = rule[field]
            group_id = self._get_rule_ips_group_id(rule['id'], direction_text)
            self._validate_cidr(group_ips, rule['id'])
            expr = self.nsxpolicy.group.build_ip_address_expression(
                [group_ips])
            tags = self.nsxpolicy.build_v3_tags_payload(
                rule, resource_type='os-neutron-fwrule-id',
                project_name=project_id)
            tags = nsxlib_utils.add_v3_tag(tags, ROUTER_FW_TAG, router_id)
            self.nsxpolicy.group.create_or_overwrite_with_conditions(
                "FW_rule_%s_%s" % (rule['id'], direction_text),
                policy_constants.DEFAULT_DOMAIN, group_id=group_id,
                description='%s: %s' % (direction_text, group_ips),
                conditions=[expr], tags=tags)
            return group_id

    def _create_network_group(self, router_id, neutron_net_id):
        scope_and_tag = "%s|%s" % ('os-neutron-net-id', neutron_net_id)
        tags = []
        tags = nsxlib_utils.add_v3_tag(tags, ROUTER_FW_TAG, router_id)
        expr = self.nsxpolicy.group.build_condition(
            cond_val=scope_and_tag,
            cond_key=policy_constants.CONDITION_KEY_TAG,
            cond_member_type=nsx_constants.TARGET_TYPE_LOGICAL_SWITCH)
        group_id = '%s-%s' % (router_id, neutron_net_id)
        self.nsxpolicy.group.create_or_overwrite_with_conditions(
            "Segment_%s" % neutron_net_id,
            policy_constants.DEFAULT_DOMAIN,
            group_id=group_id,
            description='Group for segment %s' % neutron_net_id,
            conditions=[expr],
            tags=tags)
        return group_id

    def _translate_rules(self, project_id, router_id, segment_group,
                         fwaas_rules, is_ingress, logged=False):
        """Translate a list of FWaaS rules to NSX rule structure"""
        translated_rules = []
        for rule in fwaas_rules:
            if not rule['enabled']:
                # skip disabled rules
                continue

            # Make sure the rule has a name, and it starts with the prefix
            # (backend max name length is 255)
            if rule.get('name'):
                rule_name = RULE_NAME_PREFIX + rule['name']
            else:
                rule_name = RULE_NAME_PREFIX + rule['id']
            rule_name = rule_name[:255]

            # Set rule ID with a random suffix
            rule_id = self._get_random_rule_id(rule['id'])

            action = v3_utils.translate_fw_rule_action(
                rule['action'], rule['id'])
            if not action:
                raise exceptions.FirewallInternalDriverError(
                    driver=self.internal_driver.driver_name)

            src_group = self._get_rule_cidr_group(
                project_id, router_id, rule, is_source=True,
                is_ingress=is_ingress)
            if not is_ingress and not src_group:
                src_group = segment_group
            dest_group = self._get_rule_cidr_group(
                project_id, router_id, rule, is_source=False,
                is_ingress=is_ingress)
            if is_ingress and not dest_group:
                dest_group = segment_group

            srv_id = self._translate_service(project_id, router_id, rule)
            direction = nsx_constants.IN if is_ingress else nsx_constants.OUT
            ip_protocol = (nsx_constants.IPV4 if rule.get('ip_version', 4) == 4
                           else nsx_constants.IPV6)
            rule_entry = self.nsxpolicy.gateway_policy.build_entry(
                rule_name,
                policy_constants.DEFAULT_DOMAIN,
                router_id, rule_id,
                description=rule.get('description'),
                action=action,
                source_groups=[src_group] if src_group else None,
                dest_groups=[dest_group] if dest_group else None,
                service_ids=[srv_id] if srv_id else None,
                ip_protocol=ip_protocol,
                logged=logged,
                scope=[self.nsxpolicy.tier1.get_path(router_id)],
                direction=direction)
            translated_rules.append(rule_entry)
        return translated_rules

    def _get_port_translated_rules(self, project_id, router_id, neutron_net_id,
                                   firewall_group, plugin_rules):
        """Return the list of translated FWaaS rules per port
        Add the egress/ingress rules of this port +
        default drop rules in each direction for this port.
        """
        net_group_id = self._create_network_group(
            router_id, neutron_net_id)
        port_rules = []
        # Add the firewall group ingress/egress rules only if the fw is up
        if firewall_group['admin_state_up']:
            port_rules.extend(self._translate_rules(
                project_id, router_id, net_group_id,
                firewall_group['ingress_rule_list'], is_ingress=True))
            port_rules.extend(self._translate_rules(
                project_id, router_id, net_group_id,
                firewall_group['egress_rule_list'], is_ingress=False))

        # Add the per-port plugin rules
        if plugin_rules and isinstance(plugin_rules, list):
            port_rules.extend(plugin_rules)

        # Add ingress/egress block rules for this port
        port_rules.extend([
            self.nsxpolicy.gateway_policy.build_entry(
                "Block port ingress",
                policy_constants.DEFAULT_DOMAIN, router_id,
                self._get_random_rule_id(
                    DEFAULT_RULE_ID + neutron_net_id + 'ingress'),
                action=nsx_constants.FW_ACTION_DROP,
                dest_groups=[net_group_id],
                scope=[self.nsxpolicy.tier1.get_path(router_id)],
                direction=nsx_constants.IN),
            self.nsxpolicy.gateway_policy.build_entry(
                "Block port egress",
                policy_constants.DEFAULT_DOMAIN, router_id,
                self._get_random_rule_id(
                    DEFAULT_RULE_ID + neutron_net_id + 'egress'),
                action=nsx_constants.FW_ACTION_DROP,
                scope=[self.nsxpolicy.tier1.get_path(router_id)],
                source_groups=[net_group_id],
                direction=nsx_constants.OUT)])

        return port_rules

    def _set_rules_order(self, fw_rules):
        # TODO(asarfaty): Consider adding vmware-nsxlib api for this
        # add sequence numbers to keep rules in order
        seq_num = 0
        for rule in fw_rules:
            rule.attrs['sequence_number'] = seq_num
            seq_num += 1

    def update_router_firewall(self, context, router_id, router,
                               router_interfaces, called_from_fw=False):
        """Rewrite all the FWaaS v2 rules in the router edge firewall

        This method should be called on FWaaS updates, and on router
        interfaces changes.
        The purpose of called_from_fw is to differ between fw calls and other
        router calls, and if it is True - add the service router accordingly.
        """
        plugin = self.core_plugin
        project_id = router['project_id']
        fw_rules = []
        router_with_fw = False
        # Add firewall rules per port attached to a firewall group
        for port in router_interfaces:

            # Check if this port has a firewall
            fwg = self.get_port_fwg(context, port['id'])
            if fwg:
                router_with_fw = True

                # Add plugin additional allow rules
                plugin_rules = self.core_plugin.get_extra_fw_rules(
                    context, router_id, port['id'])

                # Add the FWaaS rules for this port:ingress/egress firewall
                # rules + default ingress/egress drop rule for this port
                fw_rules.extend(self._get_port_translated_rules(
                    project_id, router_id, port['network_id'], fwg,
                    plugin_rules))

        # Add a default allow-all rule to all other traffic & ports
        fw_rules.append(self._get_default_backend_rule(router_id))
        self._set_rules_order(fw_rules)

        # Update the backend router firewall
        sr_exists_on_backend = plugin.verify_sr_at_backend(router_id)
        if called_from_fw:
            # FW action required
            if router_with_fw:
                # Firewall needed and no NSX service router: create it.
                if not sr_exists_on_backend:
                    plugin.create_service_router(
                        context, router_id, update_firewall=False)
                    sr_exists_on_backend = True
            else:
                # First, check if other services exist and use the sr
                router_with_services = plugin.service_router_has_services(
                    context, router_id, router=router)
                if not router_with_services and sr_exists_on_backend:
                    # No other services that require service router: delete it
                    # This also deleted the gateway policy.
                    self.core_plugin.delete_service_router(router_id)
                    sr_exists_on_backend = False

        if sr_exists_on_backend:
            if router_with_fw:
                self.create_or_update_router_gateway_policy(context, router_id,
                                                            router, fw_rules)
            else:
                # Do all the cleanup once the router has no more FW rules
                # create or update the edge firewall
                # TODO(asarfaty): Consider keeping the FW with default allow
                # rule instead of deletion as it may be created again soon
                self.delete_router_gateway_policy(router_id)
                self.cleanup_router_fw_resources(router_id)

    def create_or_update_router_gateway_policy(self, context, router_id,
                                               router, fw_rules):
        """Create/Overwrite gateway policy for a router with firewall rules"""
        # Check if the gateway policy already exists
        try:
            self.nsxpolicy.gateway_policy.get(policy_constants.DEFAULT_DOMAIN,
                                              map_id=router_id, silent=True)
        except nsx_lib_exc.ResourceNotFound:
            LOG.info("Going to create gateway policy for router %s", router_id)
        else:
            # only update the rules of this policy
            self.nsxpolicy.gateway_policy.update_entries(
                policy_constants.DEFAULT_DOMAIN, router_id, fw_rules,
                category=policy_constants.CATEGORY_LOCAL_GW)
            return

        tags = self.nsxpolicy.build_v3_tags_payload(
            router, resource_type='os-neutron-router-id',
            project_name=context.tenant_name)
        policy_name = GATEWAY_POLICY_NAME % router_id
        self.nsxpolicy.gateway_policy.create_with_entries(
            policy_name, policy_constants.DEFAULT_DOMAIN,
            map_id=router_id,
            description=policy_name,
            tags=tags,
            entries=fw_rules,
            category=policy_constants.CATEGORY_LOCAL_GW)

    def delete_router_gateway_policy(self, router_id):
        """Delete the gateway policy associated with a router, it it exists.
        Should be called when the router is deleted / FW removed from it
        """
        try:
            self.nsxpolicy.gateway_policy.get(policy_constants.DEFAULT_DOMAIN,
                                              map_id=router_id, silent=True)
        except nsx_lib_exc.ResourceNotFound:
            return
        self.nsxpolicy.gateway_policy.delete(policy_constants.DEFAULT_DOMAIN,
                                             map_id=router_id)

        # Also delete all groups & services
        self.cleanup_router_fw_resources(router_id)

    def cleanup_router_fw_resources(self, router_id):
        tags_to_search = [{'scope': ROUTER_FW_TAG, 'tag': router_id}]
        # Delete per rule & per network groups
        groups = self.nsxpolicy.search_by_tags(
            tags_to_search,
            self.nsxpolicy.group.entry_def.resource_type())['results']
        for group in groups:
            self.nsxpolicy.group.delete(policy_constants.DEFAULT_DOMAIN,
                                        group['id'])

        services = self.nsxpolicy.search_by_tags(
            tags_to_search,
            self.nsxpolicy.service.parent_entry_def.resource_type())['results']
        for srv in services:
            self.nsxpolicy.service.delete(srv['id'])



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers

from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import constants as plugin_const
from neutron_lib.plugins import directory

from vmware_nsx.extensions import projectpluginmap


class LBaaSNSXObjectManagerWrapper(object):
    """Wrapper class to connect the LB api with the NSX-V/V3 implementations

    This class will call the actual NSX-V LBaaS logic after translating
    the LB object into a dictionary, and will also handle success/failure cases
    """
    _core_plugin = None

    @log_helpers.log_method_call
    def __init__(self, object_type, implementor, translator, get_completor):
        super(LBaaSNSXObjectManagerWrapper, self).__init__()
        self.object_type = object_type
        self.implementor = implementor
        self.translator = translator
        self.get_completor = get_completor

    def _get_plugin(self, plugin_type):
        return directory.get_plugin(plugin_type)

    @property
    def core_plugin(self):
        if not self._core_plugin:
            self._core_plugin = (
                self._get_plugin(plugin_const.CORE))
            if self._core_plugin.is_tvd_plugin():
                # get the plugin that match this driver
                self._core_plugin = self._core_plugin.get_plugin_by_type(
                    projectpluginmap.NsxPlugins.NSX_T)
        return self._core_plugin

    def get_completor_func(self, context, obj, delete=False):
        # return a method that will be called on success/failure completion
        def completor_func(success=True):
            completor = self.get_completor()
            if completor:
                if success:
                    return completor.successful_completion(
                        context, obj, delete=delete)
                else:
                    return completor.failed_completion(
                        context, obj)

        return completor_func

    @log_helpers.log_method_call
    def create(self, context, obj, **args):
        obj_dict = self.translator(obj)
        completor_func = self.get_completor_func(context, obj)
        return self.implementor.create(context, obj_dict, completor_func,
                                       **args)

    @log_helpers.log_method_call
    def update(self, context, old_obj, new_obj, **args):
        old_obj_dict = self.translator(old_obj)
        new_obj_dict = self.translator(new_obj)
        completor_func = self.get_completor_func(context, new_obj)
        return self.implementor.update(context, old_obj_dict, new_obj_dict,
                                       completor_func, **args)

    @log_helpers.log_method_call
    def delete(self, context, obj, **args):
        obj_dict = self.translator(obj)
        completor_func = self.get_completor_func(context, obj, delete=True)
        return self.implementor.delete(context, obj_dict, completor_func,
                                       **args)

    @log_helpers.log_method_call
    def refresh(self, context, obj):
        # verify that this api exists (supported only for loadbalancer)
        if not hasattr(self.implementor, 'refresh'):
            msg = (_("LBaaS object %s does not support refresh api") %
                   self.object_type)
            raise n_exc.BadRequest(resource='edge', msg=msg)
        obj_dict = self.translator(obj)
        return self.implementor.refresh(context, obj_dict)

    @log_helpers.log_method_call
    def stats(self, context, obj):
        # verify that this api exists (supported only for loadbalancer)
        if not hasattr(self.implementor, 'stats'):
            msg = (_("LBaaS object %s does not support stats api") %
                   self.object_type)
            raise n_exc.BadRequest(resource='edge', msg=msg)
        obj_dict = self.translator(obj)
        return self.implementor.stats(context, obj_dict)

    @log_helpers.log_method_call
    def get_operating_status(self, context, id, **args):
        # verify that this api exists (supported only for loadbalancer)
        if not hasattr(self.implementor, 'get_operating_status'):
            msg = (_("LBaaS object %s does not support get_operating_status "
                     "api") % self.object_type)
            raise n_exc.BadRequest(resource='edge', msg=msg)
        return self.implementor.get_operating_status(context, id, **args)



# Copyright (c) 2016 VMware, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from neutron.tests import base

from neutron_lib import context
from oslo_utils import importutils

from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.services.trunk.nsx_p import driver as trunk_driver
from vmware_nsx.tests.unit.nsx_p import test_plugin as test_nsx_p_plugin

PLUGIN_NAME = 'vmware_nsx.plugins.nsx_p.plugin.NsxPolicyPlugin'


class TestNsxpTrunkHandler(test_nsx_p_plugin.NsxPPluginTestCaseMixin,
                           base.BaseTestCase):

    def _get_port_compute_tags_and_net(self, context, port_id):
        return True, 'net_' + port_id[-1:], []

    def setUp(self):
        super(TestNsxpTrunkHandler, self).setUp()
        self.context = context.get_admin_context()
        self.core_plugin = importutils.import_object(PLUGIN_NAME)
        self.handler = trunk_driver.NsxpTrunkHandler(self.core_plugin)
        self.handler._get_port_compute_tags_and_net = mock.Mock(
            side_effect=self._get_port_compute_tags_and_net)
        self.trunk_1 = mock.Mock()
        self.trunk_1.port_id = "parent_port_1"
        self.trunk_1.id = "trunk_1_id"

        self.trunk_2 = mock.Mock()
        self.trunk_2.port_id = "parent_port_2"

        self.sub_port_a = mock.Mock()
        self.sub_port_a.segmentation_id = 40
        self.sub_port_a.trunk_id = "trunk-1"
        self.sub_port_a.port_id = "sub_port_a"
        self.sub_port_a.segmentation_type = 'vlan'

        self.sub_port_b = mock.Mock()
        self.sub_port_b.segmentation_id = 41
        self.sub_port_b.trunk_id = "trunk-2"
        self.sub_port_b.port_id = "sub_port_b"
        self.sub_port_b.segmentation_type = 'vlan'

        self.sub_port_c = mock.Mock()
        self.sub_port_c.segmentation_id = 43
        self.sub_port_c.trunk_id = "trunk-2"
        self.sub_port_c.port_id = "sub_port_c"
        self.sub_port_c.segmentation_type = 'vlan'

    def test_trunk_created(self):
        # Create trunk with no subport
        self.trunk_1.sub_ports = []
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'attach') as m_attach:
            self.handler.trunk_created(self.context, self.trunk_1)
            m_attach.assert_called_with(
                'net_1', self.trunk_1.port_id, attachment_type='PARENT',
                tags=[{'tag': self.trunk_1.id,
                       'scope': 'os-neutron-trunk-id'}],
                vif_id=self.trunk_1.port_id)

        # Create trunk with 1 subport
        self.trunk_1.sub_ports = [self.sub_port_a]
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'attach') as m_attach:
            self.handler.trunk_created(self.context, self.trunk_1)
            calls = [
                mock.call.m_attach(
                    'net_1', self.trunk_1.port_id, attachment_type='PARENT',
                    tags=[{'tag': self.trunk_1.id,
                           'scope': 'os-neutron-trunk-id'}],
                    vif_id=self.trunk_1.port_id),
                mock.call.m_attach(
                    'net_a', self.sub_port_a.port_id, 'CHILD',
                    self.sub_port_a.port_id,
                    context_id=self.trunk_1.port_id,
                    tags=[{'tag': self.sub_port_a.trunk_id,
                           'scope': 'os-neutron-trunk-id'}],
                    traffic_tag=self.sub_port_a.segmentation_id)]
            m_attach.assert_has_calls(calls, any_order=True)

        # Create trunk with multiple subports
        self.trunk_2.sub_ports = [self.sub_port_b, self.sub_port_c]
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'attach') as m_attach:
            self.handler.trunk_created(self.context, self.trunk_2)
            calls = [
                mock.call.m_attach(
                    'net_2', self.trunk_2.port_id, attachment_type='PARENT',
                    tags=[{'tag': self.trunk_2.id,
                           'scope': 'os-neutron-trunk-id'}],
                    vif_id=self.trunk_2.port_id),
                mock.call.m_attach(
                    'net_b', self.sub_port_b.port_id, 'CHILD',
                    self.sub_port_b.port_id,
                    context_id=self.trunk_2.port_id,
                    tags=[{'tag': self.sub_port_b.trunk_id,
                           'scope': 'os-neutron-trunk-id'}],
                    traffic_tag=self.sub_port_b.segmentation_id),
                mock.call.m_attach(
                    'net_c', self.sub_port_c.port_id, 'CHILD',
                    self.sub_port_c.port_id,
                    context_id=self.trunk_2.port_id,
                    tags=[{'tag': self.sub_port_c.trunk_id,
                           'scope': 'os-neutron-trunk-id'}],
                    traffic_tag=self.sub_port_c.segmentation_id)]
            m_attach.assert_has_calls(calls, any_order=True)

    def test_trunk_deleted(self):
        # Delete trunk with no subport
        self.trunk_1.sub_ports = []
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'detach') as m_detach:
            self.handler.trunk_deleted(self.context, self.trunk_1)
            m_detach.assert_called_with(
                'net_1', self.trunk_1.port_id, vif_id=self.trunk_1.port_id,
                tags=mock.ANY)

        # Delete trunk with 1 subport
        self.trunk_1.sub_ports = [self.sub_port_a]
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'detach') as m_detach:
            self.handler.trunk_deleted(self.context, self.trunk_1)
            calls = [
                mock.call.m_detach(
                    'net_1', self.trunk_1.port_id,
                    vif_id=self.trunk_1.port_id, tags=mock.ANY),
                mock.call.m_detach(
                    'net_a', self.sub_port_a.port_id,
                    vif_id=self.sub_port_a.port_id, tags=mock.ANY)]
            m_detach.assert_has_calls(calls, any_order=True)

        # Delete trunk with multiple subports
        self.trunk_2.sub_ports = [self.sub_port_b, self.sub_port_c]
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'detach') as m_detach:
            self.handler.trunk_deleted(self.context, self.trunk_2)
            calls = [
                mock.call.m_detach(
                    'net_2', self.trunk_2.port_id,
                    vif_id=self.trunk_2.port_id, tags=mock.ANY),
                mock.call.m_detach(
                    'net_b', self.sub_port_b.port_id,
                    vif_id=self.sub_port_b.port_id, tags=mock.ANY),
                mock.call.m_detach(
                    'net_c', self.sub_port_c.port_id,
                    vif_id=self.sub_port_c.port_id, tags=mock.ANY)]
            m_detach.assert_has_calls(calls, any_order=True)

    def test_subports_added(self):
        # Update trunk with no subport
        sub_ports = []
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'attach') as m_attach:
            self.handler.subports_added(self.context, self.trunk_1, sub_ports)
            m_attach.assert_not_called()

        # Update trunk with 1 subport
        sub_ports = [self.sub_port_a]
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'attach') as m_attach:
            self.handler.subports_added(self.context, self.trunk_1, sub_ports)
        m_attach.assert_called_with(
            'net_a', self.sub_port_a.port_id, 'CHILD',
            self.sub_port_a.port_id,
            context_id=self.trunk_1.port_id,
            tags=[{'tag': self.sub_port_a.trunk_id,
                   'scope': 'os-neutron-trunk-id'}],
            traffic_tag=self.sub_port_a.segmentation_id)

        # Update trunk with multiple subports
        sub_ports = [self.sub_port_b, self.sub_port_c]
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'attach') as m_attach:
            self.handler.subports_added(self.context, self.trunk_2, sub_ports)
            calls = [
                mock.call.m_attach(
                    'net_b', self.sub_port_b.port_id, 'CHILD',
                    self.sub_port_b.port_id,
                    context_id=self.trunk_2.port_id,
                    tags=[{'tag': self.sub_port_b.trunk_id,
                           'scope': 'os-neutron-trunk-id'}],
                    traffic_tag=self.sub_port_b.segmentation_id),
                mock.call.m_attach(
                    'net_c', self.sub_port_c.port_id, 'CHILD',
                    self.sub_port_c.port_id,
                    context_id=self.trunk_2.port_id,
                    tags=[{'tag': self.sub_port_c.trunk_id,
                           'scope': 'os-neutron-trunk-id'}],
                    traffic_tag=self.sub_port_c.segmentation_id)]
            m_attach.assert_has_calls(calls, any_order=True)

    def test_subports_deleted(self):
        # Update trunk to remove no subport
        sub_ports = []
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'detach') as m_detach:
            self.handler.subports_deleted(
                self.context, self.trunk_1, sub_ports)
            m_detach.assert_not_called()

        # Update trunk to remove 1 subport
        sub_ports = [self.sub_port_a]
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'detach') as m_detach:
            self.handler.subports_deleted(
                self.context, self.trunk_1, sub_ports)
            m_detach.assert_called_with(
                'net_a', self.sub_port_a.port_id,
                vif_id=self.sub_port_a.port_id, tags=mock.ANY)

        # Update trunk to remove multiple subports
        sub_ports = [self.sub_port_b, self.sub_port_c]
        with mock.patch.object(
                self.handler.plugin_driver.nsxpolicy.segment_port,
                'detach') as m_detach:
            self.handler.subports_deleted(
                self.context, self.trunk_2, sub_ports)
            calls = [
                mock.call.m_detach(
                    'net_b', self.sub_port_b.port_id,
                    vif_id=self.sub_port_b.port_id, tags=mock.ANY),
                mock.call.m_detach(
                    'net_c', self.sub_port_c.port_id,
                    vif_id=self.sub_port_c.port_id, tags=mock.ANY)]
            m_detach.assert_has_calls(calls, any_order=True)


class TestNsxpTrunkDriver(base.BaseTestCase):
    def setUp(self):
        super(TestNsxpTrunkDriver, self).setUp()

    def test_is_loaded(self):
        core_plugin = mock.Mock()
        driver = trunk_driver.NsxpTrunkDriver.create(core_plugin)
        with mock.patch.object(core_plugin, 'plugin_type',
                               return_value=projectpluginmap.NsxPlugins.NSX_P):
            self.assertTrue(driver.is_loaded)

        with mock.patch.object(core_plugin, 'plugin_type',
                               return_value=projectpluginmap.NsxPlugins.NSX_T):
            self.assertFalse(driver.is_loaded)



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import socket

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_messaging.rpc import dispatcher
import pecan
from stevedore import driver as stevedore_driver

from octavia.api.drivers import exceptions
from octavia.api.drivers import utils as oct_utils
from octavia.db import api as db_apis
from octavia.db import repositories
from octavia_lib.api.drivers import driver_lib

from octavia_lib.api.drivers import provider_base as driver_base

from vmware_nsx.services.lbaas.octavia import constants as d_const


LOG = logging.getLogger(__name__)
cfg.CONF.import_group('oslo_messaging', 'octavia.common.config')

TRANSPORT = None
RPC_SERVER = None


def get_transport():
    global TRANSPORT
    if not TRANSPORT:
        TRANSPORT = messaging.get_rpc_transport(cfg.CONF)
    return TRANSPORT


def get_rpc_server(target, endpoints, access_policy):
    global RPC_SERVER
    if not RPC_SERVER:
        RPC_SERVER = messaging.get_rpc_server(
            TRANSPORT, target, endpoints, executor='threading',
            access_policy=access_policy)
    return RPC_SERVER


# List of keys per object type that will not be sent to the listener
unsupported_keys = {'Loadbalancer': ['vip_qos_policy_id'],
                    'Listener': ['sni_container_refs',
                                 'insert_headers',
                                 'timeout_client_data',
                                 'timeout_member_connect',
                                 'timeout_member_data',
                                 'timeout_tcp_inspect'],
                    'HealthMonitor': ['max_retries_down'],
                    'Member': ['monitor_address', 'monitor_port']}


class NSXOctaviaDriver(driver_base.ProviderDriver):
    @log_helpers.log_method_call
    def __init__(self):
        super(NSXOctaviaDriver, self).__init__()
        self._init_rpc_messaging()
        self._init_rpc_listener()
        self._init_cert_manager()
        self.repositories = repositories.Repositories()

    @log_helpers.log_method_call
    def _init_rpc_messaging(self):
        topic = d_const.OCTAVIA_TO_DRIVER_TOPIC
        transport = get_transport()
        target = messaging.Target(topic=topic, exchange="common",
                                  namespace='control', fanout=False,
                                  version='1.0')
        self.client = messaging.RPCClient(transport, target)

    @log_helpers.log_method_call
    def _init_rpc_listener(self):
        # Initialize RPC listener
        topic = d_const.DRIVER_TO_OCTAVIA_TOPIC
        server = socket.gethostname()
        target = messaging.Target(topic=topic, server=server,
                                  exchange="common", fanout=False)
        endpoints = [NSXOctaviaDriverEndpoint()]
        access_policy = dispatcher.DefaultRPCAccessPolicy

        self.octavia_server = get_rpc_server(target, endpoints,
                                             access_policy)
        self.octavia_server.start()

    @log_helpers.log_method_call
    def _init_cert_manager(self):
        self.cert_manager = stevedore_driver.DriverManager(
            namespace='octavia.cert_manager',
            name=cfg.CONF.certificates.cert_manager,
            invoke_on_load=True).driver

    def get_obj_project_id(self, obj_type, obj_dict):
        if obj_dict.get('project_id'):
            return obj_dict['project_id']
        if obj_dict.get('tenant_id'):
            return obj_dict['tenant_id']

        # look for the project id of the attached objects
        project_id = None
        if obj_dict.get('loadbalancer_id'):
            db_lb = self.repositories.load_balancer.get(
                db_apis.get_session(), id=obj_dict['loadbalancer_id'])
            if db_lb:
                project_id = db_lb.project_id
        if not project_id and obj_dict.get('pool_id'):
            db_pool = self.repositories.pool.get(
                db_apis.get_session(), id=obj_dict['pool_id'])
            if db_pool:
                project_id = db_pool.load_balancer.project_id
        if not project_id and obj_dict.get('listener_id'):
            db_list = self.repositories.listener.get(
                db_apis.get_session(), id=obj_dict['listener_id'])
            if db_list:
                project_id = db_list.load_balancer.project_id
        if not project_id and obj_dict.get('l7policy_id'):
            db_policy = self.repositories.l7policy.get(
                db_apis.get_session(), id=obj_dict['l7policy_id'])
            if db_policy:
                if db_policy.listener:
                    db_lb = db_policy.listener.load_balancer
                elif db_policy.redirect_pool:
                    db_lb = db_policy.redirect_pool.load_balancer
                if db_lb:
                    project_id = db_lb.project_id

        if not project_id:
            LOG.warning("Could bot find the tenant id for %(type)s "
                        "%(obj)s", {'type': obj_type, 'obj': obj_dict})
        return project_id

    def _get_load_balancer_dict(self, loadbalancer_id):
        if not loadbalancer_id:
            return
        db_lb = self.repositories.load_balancer.get(
            db_apis.get_session(), id=loadbalancer_id)
        if not db_lb:
            return
        lb_dict = {'name': db_lb.name, 'id': loadbalancer_id}
        if db_lb.vip:
            lb_dict['vip_port_id'] = db_lb.vip.port_id
            lb_dict['vip_address'] = db_lb.vip.ip_address
            lb_dict['vip_port_id'] = db_lb.vip.port_id
            lb_dict['vip_network_id'] = db_lb.vip.network_id
            lb_dict['vip_subnet_id'] = db_lb.vip.subnet_id
        return lb_dict

    def _get_listener_in_pool_dict(self, pool_dict, is_update):
        if 'listener' not in pool_dict:
            if pool_dict.get('listener_id'):
                db_listener = self.repositories.listener.get(
                    db_apis.get_session(), id=pool_dict['listener_id'])
                listener_obj = oct_utils.db_listener_to_provider_listener(
                    db_listener)
                listener_dict = listener_obj.to_dict(
                    recurse=False, render_unsets=True)
                listener_dict['id'] = listener_dict['listener_id']
                listener_dict['l7_policies'] = listener_dict['l7policies']
                # Add the loadbalancer to the listener dict
                if pool_dict.get('loadbalancer_id'):
                    # Generate a loadbalancer object
                    listener_dict['loadbalancer'] = (
                        self._get_load_balancer_dict(
                            pool_dict['loadbalancer_id']))
                pool_dict['listener'] = listener_dict
                if 'listeners' not in pool_dict:
                    # multiple listeners is not really supported yet
                    pool_dict['listeners'] = [listener_dict]
            # Do not add listener in update situation, as we want to use
            # the original listener of this pool
            elif not is_update:
                pool_dict['listener'] = None
                if 'listeners' not in pool_dict:
                    pool_dict['listeners'] = []

    def _get_pool_dict(self, pool_id, is_update, parent_project_id=None):
        if not pool_id:
            return {}
        db_pool = self.repositories.pool.get(db_apis.get_session(), id=pool_id)
        if not db_pool:
            return {}
        pool_obj = oct_utils.db_pool_to_provider_pool(db_pool)
        pool_dict = pool_obj.to_dict(recurse=True, render_unsets=True)
        pool_dict['id'] = pool_id
        # Get the load balancer object
        if pool_dict.get('loadbalancer_id'):
            # Generate a loadbalancer object
            pool_dict['loadbalancer'] = self._get_load_balancer_dict(
                pool_dict['loadbalancer_id'])
        if 'listener' not in pool_dict:
            self._get_listener_in_pool_dict(pool_dict, is_update)
        # make sure this pool has a project id
        if not pool_dict.get('project_id'):
            project_id = self.get_obj_project_id('Pool', pool_dict)
            if project_id is None:
                project_id = parent_project_id
            pool_dict['tenant_id'] = pool_dict['project_id'] = project_id

        return pool_dict

    def _get_hm_dict(self, hm_id, is_update):
        if not hm_id:
            return {}
        db_hm = self.repositories.health_monitor.get(
            db_apis.get_session(), id=hm_id)
        if not db_hm:
            return {}
        hm_obj = oct_utils.db_HM_to_provider_HM(db_hm)
        hm_dict = hm_obj.to_dict(recurse=True, render_unsets=True)
        hm_dict['id'] = hm_id
        # Get the pol object
        if hm_dict.get('pool_id'):
            hm_dict['pool'] = self._get_pool_dict(
                hm_dict['pool_id'], is_update)
        return hm_dict

    def update_policy_dict(self, policy_dict, policy_obj, is_update=False):
        if policy_dict.get('listener_id'):
            db_list = self.repositories.listener.get(
                db_apis.get_session(), id=policy_dict['listener_id'])
            list_obj = oct_utils.db_listener_to_provider_listener(db_list)
            list_dict = list_obj.to_dict(recurse=True, render_unsets=True)
            list_dict['id'] = policy_dict['listener_id']
            policy_dict['listener'] = list_dict
        if policy_obj.rules:
            policy_dict['rules'] = []
            for rule in policy_obj.rules:
                if isinstance(rule, dict):
                    rule_dict = rule
                else:
                    rule_dict = rule.to_dict(recurse=False, render_unsets=True)
                rule_dict['id'] = rule_dict['l7rule_id']
                policy_dict['rules'].append(rule_dict)
        elif not is_update:
            policy_dict['rules'] = []

    def _remove_unsupported_keys(self, obj_type, obj_dict):
        for key in unsupported_keys.get(obj_type, []):
            if key in obj_dict:
                if obj_dict.get(key):
                    LOG.warning("Ignoring %(key)s:%(val)s in %(type)s as the "
                                "NSX plugin does not currently support it",
                                {'key': key, 'val': obj_dict[key],
                                 'type': obj_type})
                del obj_dict[key]

    def obj_to_dict(self, obj, is_update=False, project_id=None):
        obj_type = obj.__class__.__name__
        # create a dictionary out of the object
        render_unsets = False if is_update else True
        obj_dict = obj.to_dict(recurse=True, render_unsets=render_unsets)

        # Update the dictionary to match what the nsx driver expects
        if not project_id:
            project_id = self.get_obj_project_id(obj_type, obj_dict)
        obj_dict['tenant_id'] = obj_dict['project_id'] = project_id

        if 'id' not in obj_dict:
            obj_dict['id'] = obj_dict.get('%s_id' % obj_type.lower())

        if not obj_dict.get('name') and not is_update:
            obj_dict['name'] = ""

        self._remove_unsupported_keys(obj_type, obj_dict)

        if obj_type == 'LoadBalancer':
            # clean listeners and pools for update case:
            if 'listeners' in obj_dict:
                if is_update and not obj_dict['listeners']:
                    del obj_dict['listeners']
                else:
                    if obj_dict['listeners'] is None:
                        obj_dict['listeners'] = []
                    for listener in obj_dict['listeners']:
                        listener['id'] = listener['listener_id']
                        for policy in listener.get('l7policies', []):
                            policy['id'] = policy['l7policy_id']
                            for rule in policy.get('rules', []):
                                rule['id'] = rule['l7rule_id']
            if 'pools' in obj_dict:
                if is_update and not obj_dict['pools']:
                    del obj_dict['pools']
                else:
                    if obj_dict['pools'] is None:
                        obj_dict['pools'] = []
                    for pool in obj_dict['pools']:
                        pool['id'] = pool['pool_id']
                        for member in pool.get('members', []):
                            member['id'] = member['member_id']
                        if pool.get('healthmonitor'):
                            pool['healthmonitor'] = self._get_hm_dict(
                                pool['healthmonitor']['healthmonitor_id'],
                                is_update)
                        pool['tenant_id'] = project_id

        elif obj_type == 'Listener':
            if 'l7policies' in obj_dict:
                obj_dict['l7_policies'] = obj_dict['l7policies']
            if obj_dict.get('loadbalancer_id'):
                # Generate a loadbalancer object
                obj_dict['loadbalancer'] = self._get_load_balancer_dict(
                    obj_dict['loadbalancer_id'])
            if obj_dict.get('default_pool_id'):
                # Generate the default pool object
                obj_dict['default_pool'] = self._get_pool_dict(
                    obj_dict['default_pool_id'], is_update, project_id)
            # TODO(asarfaty): add default_tls_container_id

        elif obj_type == 'Pool':
            if 'listener' not in obj_dict:
                self._get_listener_in_pool_dict(obj_dict, is_update)
            if obj_dict.get('healthmonitor'):
                obj_dict['healthmonitor']['id'] = obj_dict[
                    'healthmonitor']['healthmonitor_id']

        elif obj_type == 'Member':
            # Get the pool object
            if obj_dict.get('pool_id'):
                obj_dict['pool'] = self._get_pool_dict(
                    obj_dict['pool_id'], is_update)
                obj_dict['loadbalancer'] = None
                if 'loadbalancer' in obj_dict['pool']:
                    obj_dict['loadbalancer'] = obj_dict['pool']['loadbalancer']
                    if not obj_dict.get('subnet_id'):
                        # Use the parent vip_subnet_id instead
                        obj_dict['subnet_id'] = obj_dict['loadbalancer'][
                            'vip_subnet_id']
            elif not is_update:
                # Do not set pool & LB if in update situation, as we want to
                # use the original data of this member
                obj_dict['pool'] = None
                obj_dict['loadbalancer'] = None

        elif obj_type == 'HealthMonitor':
            # Get the pool object
            if obj_dict.get('pool_id'):
                obj_dict['pool'] = self._get_pool_dict(
                    obj_dict['pool_id'], is_update)

        elif obj_type == 'L7Policy':
            self.update_policy_dict(obj_dict, obj, is_update=is_update)

        elif obj_type == 'L7Rule':
            # Get the L7 policy object
            if obj_dict.get('l7policy_id'):
                db_policy = self.repositories.l7policy.get(
                    db_apis.get_session(), id=obj_dict['l7policy_id'])
                policy_obj = oct_utils.db_l7policy_to_provider_l7policy(
                    db_policy)
                policy_dict = policy_obj.to_dict(
                    recurse=True, render_unsets=True)
                policy_dict['id'] = obj_dict['l7policy_id']
                self.update_policy_dict(
                    policy_dict, policy_obj, is_update=is_update)
                obj_dict['policy'] = policy_dict

        LOG.debug("Translated %(type)s to dictionary: %(obj)s",
                  {'type': obj_type, 'obj': obj_dict})
        return obj_dict

    # Load Balancer
    @log_helpers.log_method_call
    def create_vip_port(self, loadbalancer_id, project_id, vip_dictionary):
        raise exceptions.NotImplementedError()

    @log_helpers.log_method_call
    def loadbalancer_create(self, loadbalancer):
        kw = {'loadbalancer': self.obj_to_dict(loadbalancer)}
        self.client.cast({}, 'loadbalancer_create', **kw)

    @log_helpers.log_method_call
    def loadbalancer_delete(self, loadbalancer, cascade=False):
        kw = {'loadbalancer': self.obj_to_dict(loadbalancer),
              'cascade': cascade}
        self.client.cast({}, 'loadbalancer_delete', **kw)

    @log_helpers.log_method_call
    def loadbalancer_failover(self, loadbalancer_id):
        LOG.error('Loadbalancer failover is handled by platform')
        raise exceptions.NotImplementedError()

    @log_helpers.log_method_call
    def loadbalancer_update(self, old_loadbalancer, new_loadbalancer):
        old_dict = self.obj_to_dict(old_loadbalancer)
        new_dict = copy.deepcopy(old_dict)
        new_dict.update(self.obj_to_dict(
            new_loadbalancer, is_update=True,
            project_id=old_dict.get('project_id')))
        kw = {'old_loadbalancer': old_dict,
              'new_loadbalancer': new_dict}
        self.client.cast({}, 'loadbalancer_update', **kw)

    # Listener
    @log_helpers.log_method_call
    def listener_create(self, listener):
        cert = None
        dict_list = self.obj_to_dict(listener)
        if dict_list.get('tls_certificate_id'):
            context = pecan.request.context.get('octavia_context')
            cert = self.cert_manager.get_cert(context,
                                              dict_list['tls_certificate_id'])
        kw = {'listener': dict_list, 'cert': cert}
        self.client.cast({}, 'listener_create', **kw)

    @log_helpers.log_method_call
    def listener_delete(self, listener):
        kw = {'listener': self.obj_to_dict(listener)}
        self.client.cast({}, 'listener_delete', **kw)

    @log_helpers.log_method_call
    def listener_update(self, old_listener, new_listener):
        old_dict = self.obj_to_dict(old_listener)
        new_dict = copy.deepcopy(old_dict)
        new_dict.update(self.obj_to_dict(
            new_listener, is_update=True,
            project_id=old_dict.get('project_id')))
        cert = None
        if new_dict.get('tls_certificate_id'):
            context = pecan.request.context.get('octavia_context')
            cert = self.cert_manager.get_cert(context,
                                              new_dict['tls_certificate_id'])
        kw = {'old_listener': old_dict,
              'new_listener': new_dict,
              'cert': cert}
        self.client.cast({}, 'listener_update', **kw)

    # Pool
    @log_helpers.log_method_call
    def pool_create(self, pool):
        kw = {'pool': self.obj_to_dict(pool)}
        self.client.cast({}, 'pool_create', **kw)

    @log_helpers.log_method_call
    def pool_delete(self, pool):
        kw = {'pool': self.obj_to_dict(pool)}
        self.client.cast({}, 'pool_delete', **kw)

    @log_helpers.log_method_call
    def pool_update(self, old_pool, new_pool):
        old_dict = self.obj_to_dict(old_pool)
        new_dict = copy.deepcopy(old_dict)
        new_pool_dict = self.obj_to_dict(
            new_pool, is_update=True, project_id=old_dict.get('project_id'))
        new_dict.update(new_pool_dict)
        kw = {'old_pool': old_dict,
              'new_pool': new_dict}
        self.client.cast({}, 'pool_update', **kw)

    # Member
    @log_helpers.log_method_call
    def member_create(self, member):
        kw = {'member': self.obj_to_dict(member)}
        self.client.cast({}, 'member_create', **kw)

    @log_helpers.log_method_call
    def member_delete(self, member):
        kw = {'member': self.obj_to_dict(member)}
        self.client.cast({}, 'member_delete', **kw)

    @log_helpers.log_method_call
    def member_update(self, old_member, new_member):
        old_dict = self.obj_to_dict(old_member)
        new_dict = copy.deepcopy(old_dict)
        new_dict.update(self.obj_to_dict(
            new_member, is_update=True, project_id=old_dict.get('project_id')))
        kw = {'old_member': old_dict,
              'new_member': new_dict}
        self.client.cast({}, 'member_update', **kw)

    @log_helpers.log_method_call
    def member_batch_update(self, members):
        raise NotImplementedError()

    # Health Monitor
    @log_helpers.log_method_call
    def health_monitor_create(self, healthmonitor):
        kw = {'healthmonitor': self.obj_to_dict(healthmonitor)}
        self.client.cast({}, 'healthmonitor_create', **kw)

    @log_helpers.log_method_call
    def health_monitor_delete(self, healthmonitor):
        kw = {'healthmonitor': self.obj_to_dict(healthmonitor)}
        self.client.cast({}, 'healthmonitor_delete', **kw)

    @log_helpers.log_method_call
    def health_monitor_update(self, old_healthmonitor, new_healthmonitor):
        old_dict = self.obj_to_dict(old_healthmonitor)
        new_dict = copy.deepcopy(old_dict)
        new_dict.update(self.obj_to_dict(
            new_healthmonitor, is_update=True,
            project_id=old_dict.get('project_id')))
        kw = {'old_healthmonitor': old_dict,
              'new_healthmonitor': new_dict}
        self.client.cast({}, 'healthmonitor_update', **kw)

    # L7 Policy
    @log_helpers.log_method_call
    def l7policy_create(self, l7policy):
        kw = {'l7policy': self.obj_to_dict(l7policy)}
        self.client.cast({}, 'l7policy_create', **kw)

    @log_helpers.log_method_call
    def l7policy_delete(self, l7policy):
        kw = {'l7policy': self.obj_to_dict(l7policy)}
        self.client.cast({}, 'l7policy_delete', **kw)

    @log_helpers.log_method_call
    def l7policy_update(self, old_l7policy, new_l7policy):
        old_dict = self.obj_to_dict(old_l7policy)
        new_dict = copy.deepcopy(old_dict)
        new_dict.update(self.obj_to_dict(
            new_l7policy, is_update=True,
            project_id=old_dict.get('project_id')))
        kw = {'old_l7policy': old_dict,
              'new_l7policy': new_dict}
        self.client.cast({}, 'l7policy_update', **kw)

    # L7 Rule
    @log_helpers.log_method_call
    def l7rule_create(self, l7rule):
        kw = {'l7rule': self.obj_to_dict(l7rule)}
        self.client.cast({}, 'l7rule_create', **kw)

    @log_helpers.log_method_call
    def l7rule_delete(self, l7rule):
        kw = {'l7rule': self.obj_to_dict(l7rule)}
        self.client.cast({}, 'l7rule_delete', **kw)

    @log_helpers.log_method_call
    def l7rule_update(self, old_l7rule, new_l7rule):
        old_dict = self.obj_to_dict(old_l7rule)
        new_dict = copy.deepcopy(old_dict)
        new_dict.update(self.obj_to_dict(
            new_l7rule, is_update=True, project_id=old_dict.get('project_id')))
        kw = {'old_l7rule': old_dict,
              'new_l7rule': new_dict}
        self.client.cast({}, 'l7rule_update', **kw)

    # Flavor
    @log_helpers.log_method_call
    def get_supported_flavor_metadata(self):
        raise exceptions.NotImplementedError()

    @log_helpers.log_method_call
    def validate_flavor(self, flavor_metadata):
        raise exceptions.NotImplementedError()


class NSXOctaviaDriverEndpoint(driver_lib.DriverLibrary):
    target = messaging.Target(namespace="control", version='1.0')

    @log_helpers.log_method_call
    def update_loadbalancer_status(self, ctxt, status):
        # refresh the driver lib session
        self.db_session = db_apis.get_session()
        try:
            return super(NSXOctaviaDriverEndpoint,
                         self).update_loadbalancer_status(status)
        except exceptions.UpdateStatusError as e:
            LOG.error("Failed to update Octavia loadbalancer status. "
                      "Status %s, Error %s", status, e.fault_string)

    @log_helpers.log_method_call
    def update_listener_statistics(self, ctxt, statistics):
        # refresh the driver lib session
        self.db_session = db_apis.get_session()
        try:
            return super(NSXOctaviaDriverEndpoint,
                         self).update_listener_statistics(statistics)
        except exceptions.UpdateStatisticsError as e:
            LOG.error("Failed to update Octavia listener statistics. "
                      "Stats %s, Error %s", statistics, e.fault_string)
